import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)
import numpy as np
import pandas as pd

from scipy import stats
from collections import defaultdict
from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix

REQUIRED_KEYS = frozenset(["feature", "low_value_high_score"])


class PercentileRankOneFeature:
    def __init__(self, feature, low_value_high_score=None, descend=None, random_state=42):
        logger.warning("DEPRECATION WARNING: PercentileRankOneFeature is being replaced by "
            "BaselineRankMultiFeature. Note, however, that the scores returned by the new "
            "ranker cannot be interpreted as percentiles."
        )
        if descend is not None:
            # If the deprecated `descend` parameter has been specified, raise a
            # warning, then use this value for low_value_high_score, which has
            # the same behavior
            logger.warning("DEPRECATION WARNING: parameter `descend` is deprecated for "
                "PercentileRankOneFeature. Use `low_value_high_score` instead."
            )
            if low_value_high_score is not None:
                raise ValueError("Only one of `descend` or `low_value_high_score` can be "
                    "specified for PercentileRankOneFeature."
                    )
            low_value_high_score = descend

        # set default this way so we can check if both have been specified above
        if low_value_high_score is None:
            low_value_high_score = False

        self.random_state = random_state
        self.feature = feature  # which feature to rank on
        self.low_value_high_score = (
            low_value_high_score
        )  # should feature be ranked so lower values -> higher scores
        self.feature_importances_ = None

    def _set_feature_importances_(self, x):
        """ Assigns feature importances following the rule: 1 for the feature we
        are ranking on, 0 for all other features.
        """
        feature_importances = [0] * len(x.columns)
        try:
            position = x.columns.get_loc(self.feature)
        except KeyError:
            raise BaselineFeatureNotInMatrix(
                (
                    "Trying to rank on a feature ({feature_name})"
                    " not included in the training matrix!".format(
                        feature_name=self.feature
                    )
                )
            )
        feature_importances[position] = 1
        self.feature_importances_ = np.array(feature_importances)

    def fit(self, x, y):
        """ Set feature importances and return self.
        """
        self._set_feature_importances_(x)
        return self

    def predict_proba(self, x):
        """ Generate the rank percentile scores and return these.
        """
        # reduce x to the selected feature, raise error if not found
        x = x[self.feature]

        # we need different behavior depending on rank ordering. percentiles
        # should be able to be interpreted as "proportion of entities ranking
        # BELOW this entity's value". scipy will assign lower ranks to lower
        # values of the feature. so if the entities have values [0, 0, 1, 2, 2],
        # the first two entities will have the lowest ranks (and therefore the
        # lowest risk scores) and the last two will have the highest ranks (and
        # highest risk scores). for the "low_value_high_score" method, we need to reverse
        # this, and for both sorting directions, we need to convert the ranks to
        # percentiles.

        # when ascending: tied entities should get the *lowest* rank, so for
        # [0, 0, 1, 2, 2] the ranks should be [1, 1, 3, 4, 4]. these can be
        # converted to the number of entities below each value by subtracting 1
        # from each rank, yielding [0, 0, 2, 3, 3]. from here, we can calculate
        # the proportions by dividing by the length of each list.
        method = "min"
        subtract = 1

        # when `low_value_high_score=True`: tied entities should get the *highest* rank, so for
        # [0, 0, 1, 2, 2] the ranks should be [2, 2, 3, 5, 5]. if we reverse
        # these ranks by substracting all items from the maximum rank (5), we
        # end up with the correct ranks for calculating percentiles:
        # [3, 3, 2, 0, 0]. to simplify the code, we first divide by the length
        # of the list then subtract the result from the maxmimum percentile (1).
        # it produces the same result as subtracting from 5 then dividing:
        #   ([5, 5, 5, 5, 5] -  [2, 2, 3, 5, 5]) / 5  = [0.6, 0.6, 0.4, 0, 0]
        # and
        #    [1, 1, 1, 1, 1] - ([2, 2, 3, 5, 5]  / 5) = [0.6, 0.6, 0.4, 0, 0]
        if self.low_value_high_score:
            method = "max"
            subtract = 0

        # get the ranks and convert to percentiles
        ranks = stats.rankdata(x, method)
        ranks = [(rank - subtract) / len(x) for rank in ranks]
        if self.low_value_high_score:
            ranks = [1 - rank for rank in ranks]

        # format it like sklearn output and return
        return np.array([np.zeros(len(x)), ranks]).transpose()


class BaselineRankMultiFeature:
    def __init__(self, rules, random_state=42):
        if not isinstance(rules, list):
            rules = [rules]

        # validate rules: must have feature and sort order
        for rule in rules:
            if not isinstance(rule, dict):
                raise ValueError('Rules for BaselineRankMultiFeature must be of type dict')
            if not rule.keys() >= REQUIRED_KEYS:
                raise ValueError(f'BaselineRankMultiFeature rule "{rule}" missing one or more required keys ({REQUIRED_KEYS})')

        self.rules = rules
        self.feature_importances_ = None

    @property
    def all_feature_names(self):
        return [rule["feature"] for rule in self.rules]

    @property
    def all_sort_directions(self):
        # note that ascending=True sort will mean low values get low scores,
        # so negate the parameter direction to get the right relationship
        return [not rule['low_value_high_score'] for rule in self.rules]

    def _set_feature_importances_(self, x):
        """ Assigns feature importances following the rule: 1 for the features
        we are thresholding on, 0 for all other features.
        """
        feature_importances = [0] * len(x.columns)
        for feature_name in self.all_feature_names:
            try:
                position = x.columns.get_loc(feature_name)
            except KeyError:
                raise BaselineFeatureNotInMatrix(
                    (
                        "Rules refer to a feature ({feature_name}) not included in "
                        "the training matrix!".format(feature_name=feature_name)
                    )
                )
            feature_importances[position] = 1
        self.feature_importances_ = np.array(feature_importances)

    def fit(self, x, y):
        """ Set feature importances and return self.
        """
        self._set_feature_importances_(x)
        return self

    def predict_proba(self, x):
        """ Generate the rank scores and return these.
        """
        # reduce x to the selected set of features
        x = x[self.all_feature_names].reset_index(drop=True)

        x = x.sort_values(self.all_feature_names, ascending=self.all_sort_directions)

        # initialize curr_rank to -1 so the first record will have rank 0 (hence "score"
        # will range from 0 to 1)
        ranks = []
        curr_rank = -1
        prev = []

        # calculate ranks over sorted records, giving ties the same rank
        for rec in x.values:
            if not np.array_equal(prev, rec):
                curr_rank += 1
            ranks.append(curr_rank)
            prev = rec

        # normalize to 0 to 1 range
        max_rank = ranks[-1]
        x['score'] = [r/max_rank for r in ranks]

        # reset back to original sort order, calculate "score" for "0 class"
        scores_1 = x.sort_index()['score'].values
        scores_0 = np.array([1-s for s in scores_1])

        return np.array([scores_0, scores_1]).transpose()


class LinearRanker:
    """This baseline objects ranks importance based on a linear combination
    given by self.weights * self.features.
    For example, if self.weights=[0.20, 0.80] and self.features=['A', 'B'] then this
    baseline creates a score based on the linear combination and ranks people accordingly

    Args:
        features (list): list of features to rank on. features[0] and features[-1] are the
           most and least important features
        weights (list): list of weights to use in the linear combination

    Returns:
        scores (array): Numpy array of shape (n, 2) where n is the number of rows in X. 
    """
    def __init__(self, features, weights, random_state=42):
        self.features = features
        self.weights = np.array(weights) / sum(np.array(weights))
        #self.__name__ = 'LinearRanker'
        self.feature_importances_ = None
        self.random_state = random_state
    
    def _set_feature_importance(self, x):
        """ Assigns feature importances based on the weights provided.
        """
        diff_features = set(self.features).difference(set(x.columns.tolist()))
        if len(diff_features) > 0:
            for feature in list(diff_features): 
                logger.error(f"LinearRanker refers to feature {feature} not included in the training matrix!")
                raise Exception(f"LinearRanker refers to feature {feature} not included in the training matrix!")
        
        # feature importance 
        df = pd.DataFrame({'feature': self.features, 'weight': self.weights})
        self.feature_importances_ = np.array(df.weight) 
       

    def fit(self, x, y):
        """Run sanity checks and populate self.feature_importances_
        """
        # Ensure all the features requested are real features
        self._set_feature_importance(x) 

        return self

    def predict_proba(self, x):
        """Returns an n by 2 matrix. The first col is 0 and the second one is
        the linear combination of self.weights and self.features.

        Args:
        x: dataframe with multiindex=(joid, as_of_date) and n rows i.e., joid-date pairs

        Returns:
        np.array of shape (n, 2) where n is the number of rows in X.

        This output has this structure to mimic the behavior of sklearn.model.predic_proba()
        """ 

        # Change scale of each column from 0 to 1
        def index_build(feat):
            perc_99 = x[feat].quantile(0.99)
            x.loc[x[feat] > perc_99, feat] = perc_99
            # Apply function to non-zero percentile columns
            if perc_99 != 0:
                x[feat] = x[feat]/perc_99
        
        for feature in self.features:
            index_build(feature)

        # Compute the score as a linear combination 
        score = np.array((self.weights * x[self.features]).sum(axis=1))
        rv = np.array([1-score, score]).T

        return rv


from triage.component.results_schema import RetrainModel, Retrain
from triage.component.catwalk.utils import db_retry, filename_friendly_hash

import re
from sqlalchemy.orm import sessionmaker
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)


def experiment_config_from_model_id(db_engine, model_id):
    """Get original experiment config from model_id 
    Args:
            db_engine (sqlalchemy.db.engine)
            model_id (int) The id of a given model in the database

    Returns: (dict) experiment config
    """
    get_experiment_query = '''select experiments.config
    from triage_metadata.triage_runs
    join triage_metadata.models on (triage_runs.id = models.built_in_triage_run)
    join triage_metadata.experiments 
        on (experiments.experiment_hash = triage_runs.run_hash and triage_runs.run_type='experiment')
    where model_id = %s
    '''
    (config,) = db_engine.execute(get_experiment_query, model_id).first()
    return config


def experiment_config_from_model_group_id(db_engine, model_group_id):
    """Get original experiment config from model_id 
    Args:
            db_engine (sqlalchemy.db.engine)
            model_id (int) The id of a given model in the database

    Returns: (dict) experiment config
    """
    get_experiment_query = '''
    select triage_runs.id as run_id, experiments.config
    from triage_metadata.triage_runs
    join triage_metadata.models
    on (triage_runs.id = models.built_in_triage_run)
    join triage_metadata.experiments
    on (experiments.experiment_hash = triage_runs.run_hash and triage_runs.run_type='experiment')
    where model_group_id = %s
    order by triage_runs.start_time desc
    '''
    (run_id, config) = db_engine.execute(get_experiment_query, model_group_id).first()
    return run_id, config


def get_model_group_info(db_engine, model_group_id):
    query = """
    SELECT model_group_id, model_type, hyperparameters, model_id as model_id_last_split
    FROM triage_metadata.models
    WHERE model_group_id = %s
    ORDER BY train_end_time DESC
    """
    model_group_info = db_engine.execute(query, model_group_id).fetchone()
    return dict(model_group_info)


def train_matrix_info_from_model_id(db_engine, model_id):
    """Get original train matrix information from model_id 
    Args:
            db_engine (sqlalchemy.db.engine)
            model_id (int) The id of a given model in the database

    Returns: (str, dict) matrix uuid and matrix metadata
    """
    get_train_matrix_query = """
        select matrix_uuid, matrices.matrix_metadata
        from triage_metadata.matrices
        join triage_metadata.models on (models.train_matrix_uuid = matrices.matrix_uuid)
        where model_id = %s
    """
    return db_engine.execute(get_train_matrix_query, model_id).first()


def test_matrix_info_from_model_id(db_engine, model_id):
    """Get original test matrix information from model_id 

    Note: because a model may have been tested on multiple matrices, this
          chooses the matrix associated with the most recently run experiment
          (then randomly if multiple test matrices are associated with the model_id
          in that experiment). Generally, this will be an edge case, but may be
          worth considering providing more control over which to choose here.
    
    Args:
            db_engine (sqlalchemy.db.engine)
            model_id (int) The id of a given model in the database

    Returns: (str, dict) matrix uuid and matrix metadata
    """
    get_test_matrix_query = """
        select mat.matrix_uuid, mat.matrix_metadata
        from triage_metadata.matrices mat
        join test_results.prediction_metadata pm on (pm.matrix_uuid = mat.matrix_uuid)
        join triage_metadata.triage_runs tr
            on (mat.built_by_experiment = tr.run_hash AND tr.run_type='experiment')
        where pm.model_id = %s
        order by start_time DESC, RANDOM()
        limit 1
    """
    return db_engine.execute(get_test_matrix_query, model_id).first()



def temporal_params_from_matrix_metadata(db_engine, model_id):
    """Read temporal parameters associated with model training/testing from the associated
    matrices. Because a grid of multiple values may be provided in the experiment config
    for these parameters, we need to find the specific values that were actually used for
    the given model at runtime.

    Args:
            db_engine (sqlalchemy.db.engine)
            model_id (int) The id of a given model in the database

    Returns: (dict) The parameters for use in a temporal config for timechop
    """
    train_uuid, train_metadata = train_matrix_info_from_model_id(db_engine, model_id)
    test_uuid, test_metadata = test_matrix_info_from_model_id(db_engine, model_id)

    temporal_params = {}

    temporal_params['training_as_of_date_frequencies'] = train_metadata['training_as_of_date_frequency']
    temporal_params['test_as_of_date_frequencies'] = test_metadata['test_as_of_date_frequency']
    temporal_params['max_training_histories'] = [ train_metadata['max_training_history'] ]
    temporal_params['test_durations'] = [ test_metadata['test_duration'] ]
    temporal_params['training_label_timespans'] = [ train_metadata.get('training_label_timespan', train_metadata['label_timespan']) ]
    temporal_params['test_label_timespans'] = [ test_metadata.get('test_label_timespan', test_metadata['label_timespan']) ]

    return temporal_params

def get_feature_names(aggregation, matrix_metadata):
    """Returns a feature group name and a list of feature names from a SpacetimeAggregation object"""
    feature_prefix = aggregation.prefix
    logger.spam("Feature prefix = %s", feature_prefix)
    feature_group = aggregation.get_table_name(imputed=True).split('.')[1].replace('"', '')
    logger.spam("Feature group = %s", feature_group)
    feature_names_in_group = [f for f in matrix_metadata['feature_names'] if re.match(f'\\A{feature_prefix}_entity_id', f)]
    logger.spam("Feature names in group = %s", feature_names_in_group)
    
    return feature_group, feature_names_in_group


def get_feature_needs_imputation_in_train(aggregation, feature_names):
    """Returns features that needs imputation from training data
    Args:
        aggregation (SpacetimeAggregation)
        feature_names (list) A list of feature names
    """
    features_imputed_in_train = [
        f for f in set(feature_names)
        if not f.endswith('_imp') 
        and aggregation.imputation_flag_base(f) + '_imp' in feature_names
    ]
    logger.spam("Features imputed in train = %s", features_imputed_in_train)
    return features_imputed_in_train


def get_feature_needs_imputation_in_production(aggregation, db_engine):
    """Returns features that needs imputation from triage_production
    Args:
        aggregation (SpacetimeAggregation)
        db_engine (sqlalchemy.db.engine)
    """
    with db_engine.begin() as conn:
        nulls_results = conn.execute(aggregation.find_nulls())
    
    null_counts = nulls_results.first().items()
    features_imputed_in_production = [col for (col, val) in null_counts if val is not None and val > 0]
    
    return features_imputed_in_production


def get_retrain_config_from_model_id(db_engine, model_id):
    query = """
    SELECT re.config FROM triage_metadata.models m
    LEFT JOIN triage_metadata.triage_runs r 
        ON m.built_in_triage_run = r.id 
    LEFT JOIN triage_metadata.retrain re 
        ON (re.retrain_hash = r.run_hash and r.run_type='retrain')
    WHERE m.model_id = %s;
    """

    (config,) = db_engine.execute(query, model_id).first()
    return config


@db_retry
def associate_models_with_retrain(retrain_hash, model_hashes, db_engine):
    session = sessionmaker(bind=db_engine)()
    for model_hash in model_hashes:
        session.merge(RetrainModel(retrain_hash=retrain_hash, model_hash=model_hash))
    session.commit()
    session.close()
    logger.spam("Associated models with retrain in database")

@db_retry
def save_retrain_and_get_hash(config, db_engine):
    retrain_hash = filename_friendly_hash(config)
    session = sessionmaker(bind=db_engine)()
    session.merge(Retrain(retrain_hash=retrain_hash, config=config))
    session.commit()
    session.close()
    return retrain_hash


def cohort_config_from_label_config(label_config):
    """Hande the cases where the cohort query is not specified""" 
    
    label_query = label_config['query']
    
    cohort_config = dict()
    cohort_config['name'] = 'default'

    # We can't have the label_timespan in the cohort query
    label_query = label_query.replace('{label_timespan}', '1week')

    # We use the label query as a subquery and extract the entity ids
    cohort_config['query'] = f"""
        select 
            entity_id
        from ({label_query}) as lq
    """

    return cohort_config



from functools import partial
import pandas as pd
import numpy as np

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

def downcast_matrix(df):
    """Downcast the numeric values of a matrix.

    This will make the matrix use less memory by turning, every number into
    float32. It's more expensive in time to try to convert int64 into int32 
    than just convert the whole matrix in float32, which still is less memory
    intensive than the original matrix. 

    Operates on the dataframe as passed, without doing anything to the index.
    Callers may pass an index-less dataframe if they wish to re-add the index afterwards
    and save memory on the index storage.
    """
    logger.spam("Downcasting matrix.")
    logger.spam(f"Starting memory usage: {df.memory_usage(deep=True).sum()/1000000} MB")
    logger.spam(f"Initial types: \n {df.dtypes}")

    df = df.apply(lambda x: x.astype('float32'))
    
    logger.spam("Downcasting matrix completed.")
    logger.spam(f"Final memory usage: {df.memory_usage(deep=True).sum()/1000000} MB")
    logger.spam(f"Final data types: \n {df.dtypes}")

    return df



from triage.component.results_schema import upgrade_db, Retrain, TriageRun, TriageRunStatus
from triage.component.architect.entity_date_table_generators import EntityDateTableGenerator, DEFAULT_ACTIVE_STATE
from triage.component.architect.features import (
        FeatureGenerator, 
        FeatureDictionaryCreator, 
        FeatureGroupCreator,
        FeatureGroupMixer,
)
from triage.component.architect.feature_group_creator import FeatureGroup
from triage.component.architect.builders import MatrixBuilder
from triage.component.architect.planner import Planner
from triage.component.architect.label_generators import LabelGenerator
from triage.component.timechop import Timechop
from triage.component.catwalk.storage import ModelStorageEngine, ProjectStorage
from triage.component.catwalk import ModelTrainer
from triage.component.catwalk.model_trainers import flatten_grid_config
from triage.component.catwalk.predictors import Predictor
from triage.component.catwalk.utils import retrieve_model_hash_from_id, filename_friendly_hash, retrieve_experiment_seed_from_run_id
from triage.util.conf import convert_str_to_relativedelta, dt_from_str
from triage.util.db import scoped_session, get_for_update
from triage.util.introspection import classpath
from triage.tracking import (
    infer_git_hash, 
    infer_ec2_instance_type, 
    infer_installed_libraries,
    infer_python_version,
    infer_triage_version,
    infer_log_location,
    record_cohort_table_name,
    record_labels_table_name,
    record_matrix_building_started,
    record_model_building_started,
)
from .utils import (
    experiment_config_from_model_id,
    experiment_config_from_model_group_id,
    get_model_group_info,
    train_matrix_info_from_model_id,
    get_feature_names,
    get_feature_needs_imputation_in_train,
    get_feature_needs_imputation_in_production,
    associate_models_with_retrain,
    save_retrain_and_get_hash,
    get_retrain_config_from_model_id,
    temporal_params_from_matrix_metadata,
    cohort_config_from_label_config
)


from collections import OrderedDict
import json
import random
import platform
import getpass
import os
from datetime import datetime

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)



def predict_forward_with_existed_model(db_engine, project_path, model_id, as_of_date, cohort_query=None):
    """Predict forward given model_id and as_of_date and store the prediction in database

    Args:
            db_engine (sqlalchemy.db.engine)
            project_storage (catwalk.storage.ProjectStorage)
            model_id (int) The id of a given model in the database
            as_of_date (string) a date string like "YYYY-MM-DD"
    """
    logger.spam("In PREDICT LIST................")
    upgrade_db(db_engine=db_engine)
    project_storage = ProjectStorage(project_path)
    matrix_storage_engine = project_storage.matrix_storage_engine()
    
    # 1. Get feature and cohort config from database
    (train_matrix_uuid, matrix_metadata) = train_matrix_info_from_model_id(db_engine, model_id)
    experiment_config = experiment_config_from_model_id(db_engine, model_id)
 
    # 2. Generate cohort
    if experiment_config.get('cohort_config') is None:
        # If a separate cohort_config is not defined in the config
        logger.info('Experiment config does not contain a cohort config. Using the label query')
        experiment_config['cohort_config'] = cohort_config_from_label_config(experiment_config['label_config'])

    cohort_table_name = f"triage_production.cohort_{experiment_config['cohort_config']['name']}"
    cohort_table_generator = EntityDateTableGenerator(
        db_engine=db_engine,
        query=experiment_config.get('cohort_config', {}).get('query'),
        entity_date_table_name=cohort_table_name
    )
    cohort_table_generator.generate_entity_date_table(as_of_dates=[dt_from_str(as_of_date)])
    
    # 3. Generate feature aggregations
    feature_generator = FeatureGenerator(
        db_engine=db_engine,
        features_schema_name="triage_production",
        feature_start_time=experiment_config['temporal_config']['feature_start_time'],
    )
    collate_aggregations = feature_generator.aggregations(
        feature_aggregation_config=experiment_config['feature_aggregations'],
        feature_dates=[as_of_date],
        state_table=cohort_table_name
    )
    feature_generator.process_table_tasks(
        feature_generator.generate_all_table_tasks(
            collate_aggregations,
            task_type='aggregation'
        )
    )

    # 4. Reconstruct feature disctionary from feature_names and generate imputation
    
    reconstructed_feature_dict = FeatureGroup()
    imputation_table_tasks = OrderedDict()

    for aggregation in collate_aggregations:
        feature_group, feature_names = get_feature_names(aggregation, matrix_metadata)
        reconstructed_feature_dict[feature_group] = feature_names

        # Make sure that the features imputed in training should also be imputed in production
        
        features_imputed_in_train = get_feature_needs_imputation_in_train(aggregation, feature_names)
        
        features_imputed_in_production = get_feature_needs_imputation_in_production(aggregation, db_engine)

        total_impute_cols = set(features_imputed_in_production) | set(features_imputed_in_train)
        total_nonimpute_cols = set(f for f in set(feature_names) if '_imp' not in f) - total_impute_cols
        
        task_generator = feature_generator._generate_imp_table_tasks_for
        
        imputation_table_tasks.update(task_generator(
            aggregation,
            impute_cols=list(total_impute_cols),
            nonimpute_cols=list(total_nonimpute_cols)
            )
        )
    feature_generator.process_table_tasks(imputation_table_tasks)

    # 5. Build matrix
    db_config = {
        "features_schema_name": "triage_production",
        "labels_schema_name": "public",
        "cohort_table_name": cohort_table_name,
    }

    matrix_builder = MatrixBuilder(
        db_config=db_config,
        matrix_storage_engine=matrix_storage_engine,
        engine=db_engine,
        experiment_hash=None,
        replace=True,
    )
       
    feature_start_time = experiment_config['temporal_config']['feature_start_time']
    label_name = experiment_config['label_config']['name']
    label_type = 'binary'
    cohort_name = experiment_config['cohort_config']['name']
    user_metadata = experiment_config.get('user_metadata', {})
    
    # Use timechop to get the time definition for production
    temporal_config = experiment_config["temporal_config"]
    temporal_config.update(temporal_params_from_matrix_metadata(db_engine, model_id))
    timechopper = Timechop(**temporal_config)
    prod_definitions = timechopper.define_test_matrices(
            train_test_split_time=dt_from_str(as_of_date), 
            test_duration=temporal_config['test_durations'][0],
            test_label_timespan=temporal_config['test_label_timespans'][0]
    )

    last_split_definition = prod_definitions[-1]

    # formating the datetimes as strings to be saved as JSON
    last_split_definition['first_as_of_time'] = str(last_split_definition['first_as_of_time'])
    last_split_definition['last_as_of_time'] = str(last_split_definition['last_as_of_time'])
    last_split_definition['matrix_info_end_time'] = str(last_split_definition['matrix_info_end_time'])
    last_split_definition['as_of_times'] = [str(last_split_definition['as_of_times'][0])]

    matrix_metadata = Planner.make_metadata(
            last_split_definition,
            reconstructed_feature_dict,
            label_name,
            label_type,
            cohort_name,
            'production',
            feature_start_time,
            user_metadata,
    )
    
    matrix_metadata['matrix_id'] = str(as_of_date) +  f'_model_id_{model_id}' + '_risklist'

    matrix_uuid = filename_friendly_hash(matrix_metadata)
    
    matrix_builder.build_matrix(
        as_of_times=[as_of_date],
        label_name=label_name,
        label_type=label_type,
        feature_dictionary=reconstructed_feature_dict,
        matrix_metadata=matrix_metadata,
        matrix_uuid=matrix_uuid,
        matrix_type="production",
    )
    
    # 6. Predict the risk score for production
    predictor = Predictor(
        model_storage_engine=project_storage.model_storage_engine(),
        db_engine=db_engine,
        rank_order='best'
    )

    predictor.predict(
        model_id=model_id,
        matrix_store=matrix_storage_engine.get_store(matrix_uuid),
        misc_db_parameters={},
        train_matrix_columns=matrix_storage_engine.get_store(train_matrix_uuid).columns()
    )
    

class Retrainer:
    """Given a model_group_id and prediction_date, retrain a model using the all the data till prediction_date 
    Args:
        db_engine (sqlalchemy.engine)
        project_path (string)
        model_group_id (string)
    """
    def __init__(self, db_engine, project_path, model_group_id):
        self.retrain_hash = None
        self.db_engine = db_engine
        upgrade_db(db_engine=self.db_engine)
        self.project_storage = ProjectStorage(project_path)
        self.model_group_id = model_group_id
        self.model_group_info = get_model_group_info(self.db_engine, self.model_group_id)
        self.matrix_storage_engine = self.project_storage.matrix_storage_engine()
        self.triage_run_id, self.experiment_config = experiment_config_from_model_group_id(self.db_engine, self.model_group_id)

        # This feels like it needs some refactoring since in some edge cases at least the test matrix temporal parameters
        # might differ across models in the mdoel group (the training ones shouldn't), but this should probably work for
        # the vast majorty of use cases...
        self.experiment_config['temporal_config'].update(temporal_params_from_matrix_metadata(self.db_engine, self.model_group_info['model_id_last_split']))

        # Since "testing" here is predicting forward to a single new date, the test_duration should always be '0day'
        # (regardless of what it may have been before)
        self.experiment_config['temporal_config']['test_durations'] = ['0day']

        # These lists should now only contain one item (the value actually used for the last model in this group)
        self.training_label_timespan = self.experiment_config['temporal_config']['training_label_timespans'][0]
        self.test_label_timespan = self.experiment_config['temporal_config']['test_label_timespans'][0]
        self.test_duration = self.experiment_config['temporal_config']['test_durations'][0]
        self.feature_start_time=self.experiment_config['temporal_config']['feature_start_time']

        # Handling the case where a separate cohort_config is not defined
        if self.experiment_config.get('cohort_config') is None:
            self.experiment_config['cohort_config'] = cohort_config_from_label_config(self.experiment_config['label_config'])

        self.label_name = self.experiment_config['label_config']['name']

        # In case only the label config is created
        self.cohort_name = self.experiment_config.get('cohort_config', {}).get('name', 'default')
        
        # User metadata is optional
        self.user_metadata = self.experiment_config.get('user_metadata', {})

        
        self.feature_dictionary_creator = FeatureDictionaryCreator(
            features_schema_name='triage_production', db_engine=self.db_engine
        )

        self.label_generator = LabelGenerator(
            label_name=self.experiment_config['label_config'].get("name", None),
            query=self.experiment_config['label_config']["query"],
            replace=True,
            db_engine=self.db_engine,
        )        
        
        self.labels_table_name = "labels_{}_{}_production".format(
            self.experiment_config['label_config'].get('name', 'default'),
            filename_friendly_hash(self.experiment_config['label_config']['query'])
        )

        self.feature_generator = FeatureGenerator(
            db_engine=self.db_engine,
            features_schema_name="triage_production",
            feature_start_time=self.feature_start_time,
        )

        self.model_trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=ModelStorageEngine(self.project_storage),
            db_engine=self.db_engine,
            replace=True,
            run_id=self.triage_run_id,
        )
    
    def get_temporal_config_for_retrain(self, prediction_date):
        temporal_config = self.experiment_config['temporal_config'].copy()
        temporal_config['feature_end_time'] = datetime.strftime(prediction_date, "%Y-%m-%d")
        temporal_config['label_end_time'] = datetime.strftime(
                prediction_date + convert_str_to_relativedelta(self.test_label_timespan), 
                "%Y-%m-%d")
        # just needs to be bigger than the gap between the label start and end times
        # to ensure we only get one time split for the retraining
        temporal_config['model_update_frequency'] = '%syears' % (
                dt_from_str(temporal_config['label_end_time']).year - 
                dt_from_str(temporal_config['label_start_time']).year + 10
                )
        
        return temporal_config

    def generate_all_labels(self, as_of_date):
        self.label_generator.generate_all_labels(
                labels_table=self.labels_table_name, 
                as_of_dates=[as_of_date], 
                label_timespans=[self.training_label_timespan]
        )

    def generate_entity_date_table(self, as_of_date, entity_date_table_name):
        cohort_table_generator = EntityDateTableGenerator(
            db_engine=self.db_engine,
            query=self.experiment_config.get('cohort_config', {}).get('query'),
            labels_table_name=self.labels_table_name,
            entity_date_table_name=entity_date_table_name
        )
        cohort_table_generator.generate_entity_date_table(as_of_dates=[dt_from_str(as_of_date)])
       
    def get_collate_aggregations(self, as_of_date, state_table):
        collate_aggregations = self.feature_generator.aggregations(
            feature_aggregation_config=self.experiment_config['feature_aggregations'],
            feature_dates=[as_of_date],
            state_table=state_table
        )
        return collate_aggregations

    def get_feature_dict_and_imputation_task(self, collate_aggregations, model_id):
        (train_matrix_uuid, matrix_metadata) = train_matrix_info_from_model_id(self.db_engine, model_id)
        reconstructed_feature_dict = FeatureGroup()
        imputation_table_tasks = OrderedDict()
        for aggregation in collate_aggregations:
            feature_group, feature_names = get_feature_names(aggregation, matrix_metadata)
            reconstructed_feature_dict[feature_group] = feature_names
            # Make sure that the features imputed in training should also be imputed in production
            
            features_imputed_in_train = get_feature_needs_imputation_in_train(aggregation, feature_names)
            
            features_imputed_in_production = get_feature_needs_imputation_in_production(aggregation, self.db_engine)

            total_impute_cols = set(features_imputed_in_production) | set(features_imputed_in_train)
            total_nonimpute_cols = set(f for f in set(feature_names) if '_imp' not in f) - total_impute_cols
            
            task_generator = self.feature_generator._generate_imp_table_tasks_for
            
            imputation_table_tasks.update(task_generator(
                aggregation,
                impute_cols=list(total_impute_cols),
                nonimpute_cols=list(total_nonimpute_cols)
                )
            )
        return reconstructed_feature_dict, imputation_table_tasks

    def retrain(self, prediction_date):
        """Retrain a model by going back one split from prediction_date, so the as_of_date for training would be (prediction_date - training_label_timespan)
        
        Args:
            prediction_date(str) 
        """
        # Retrain config and hash
        retrain_config = {
            "model_group_id": self.model_group_id,
            "prediction_date": prediction_date,
            "test_label_timespan": self.test_label_timespan,
            "test_duration": self.test_duration,

        }
        self.retrain_hash = save_retrain_and_get_hash(retrain_config, self.db_engine)

        with get_for_update(self.db_engine, Retrain, self.retrain_hash) as retrain:
            retrain.prediction_date = prediction_date


        # Timechop
        prediction_date = dt_from_str(prediction_date)
        temporal_config = self.get_temporal_config_for_retrain(prediction_date)
        timechopper = Timechop(**temporal_config)
        chops = timechopper.chop_time()
        assert len(chops) == 1
        chops_train_matrix = chops[0]['train_matrix']
        as_of_date = datetime.strftime(chops_train_matrix['last_as_of_time'], "%Y-%m-%d")
        retrain_definition = {
            'first_as_of_time': str(chops_train_matrix['first_as_of_time']),
            'last_as_of_time': str(chops_train_matrix['last_as_of_time']),
            'matrix_info_end_time': str(chops_train_matrix['matrix_info_end_time']),
            'as_of_times': [as_of_date],
            'training_label_timespan': chops_train_matrix['training_label_timespan'],
            'max_training_history': chops_train_matrix['max_training_history'],
            'training_as_of_date_frequency': chops_train_matrix['training_as_of_date_frequency'],
        }

        # Set ExperimentRun
        run = TriageRun(
            start_time=datetime.now(),
            git_hash=infer_git_hash(),
            triage_version=infer_triage_version(),
            python_version=infer_python_version(),
            run_type="retrain",
            run_hash=self.retrain_hash,
            last_updated_time=datetime.now(),
            current_status=TriageRunStatus.started,
            installed_libraries=infer_installed_libraries(),
            platform=platform.platform(),
            os_user=getpass.getuser(),
            working_directory=os.getcwd(),
            ec2_instance_type=infer_ec2_instance_type(),
            log_location=infer_log_location(),
            experiment_class_path=classpath(self.__class__),
            random_seed = retrieve_experiment_seed_from_run_id(self.db_engine, self.triage_run_id),
        )
        run_id = None
        with scoped_session(self.db_engine) as session:
            session.add(run)
            session.commit()
            run_id = run.run_id
        if not run_id:
            raise ValueError("Failed to retrieve run_id from saved row")

        # set ModelTrainer's run_id and experiment_hash for Retrain run
        self.model_trainer.run_id = run_id
        self.model_trainer.experiment_hash = self.retrain_hash

        # 1. Generate all labels
        self.generate_all_labels(as_of_date)
        record_labels_table_name(run_id, self.db_engine, self.labels_table_name)

        # 2. Generate cohort
        cohort_table_name = f"triage_production.cohort_{self.cohort_name}_retrain"
        self.generate_entity_date_table(as_of_date, cohort_table_name)
        record_cohort_table_name(run_id, self.db_engine, cohort_table_name)

        # 3. Generate feature aggregations
        collate_aggregations = self.get_collate_aggregations(as_of_date, cohort_table_name)
        feature_aggregation_table_tasks = self.feature_generator.generate_all_table_tasks(
            collate_aggregations,
            task_type='aggregation'
        )
        self.feature_generator.process_table_tasks(feature_aggregation_table_tasks)

        # 4. Reconstruct feature disctionary from feature_names and generate imputation
        reconstructed_feature_dict, imputation_table_tasks = self.get_feature_dict_and_imputation_task(
                collate_aggregations,
                self.model_group_info['model_id_last_split'],
        )
        feature_group_creator = FeatureGroupCreator(self.experiment_config['feature_group_definition'])
        feature_group_mixer = FeatureGroupMixer(["all"])
        feature_group_dict = feature_group_mixer.generate(
            feature_group_creator.subsets(reconstructed_feature_dict)
        )[0]
        self.feature_generator.process_table_tasks(imputation_table_tasks)
        # 5. Build new matrix
        db_config = {
            "features_schema_name": "triage_production",
            "labels_schema_name": "public",
            "cohort_table_name": cohort_table_name,
            "labels_table_name": self.labels_table_name,
        }

        record_matrix_building_started(run_id, self.db_engine)
        matrix_builder = MatrixBuilder(
            db_config=db_config,
            matrix_storage_engine=self.matrix_storage_engine,
            engine=self.db_engine,
            experiment_hash=None,
            replace=True,
        )
        new_matrix_metadata = Planner.make_metadata(
            matrix_definition=retrain_definition,
            feature_dictionary=feature_group_dict,
            label_name=self.label_name,
            label_type='binary',
            cohort_name=self.cohort_name,
            matrix_type='train',
            feature_start_time=dt_from_str(self.feature_start_time),
            user_metadata=self.user_metadata,
        )
        
        new_matrix_metadata['matrix_id'] = "_".join(
            [
                self.label_name,
                'binary',
                str(as_of_date),
                'retrain',
                ]
        )

        matrix_uuid = filename_friendly_hash(new_matrix_metadata)
        matrix_builder.build_matrix(
            as_of_times=[as_of_date],
            label_name=self.label_name,
            label_type='binary',
            feature_dictionary=feature_group_dict,
            matrix_metadata=new_matrix_metadata,
            matrix_uuid=matrix_uuid,
            matrix_type="train",
        )
        retrain_model_comment = 'retrain_' + str(datetime.now())

        misc_db_parameters = {
            'train_end_time': dt_from_str(as_of_date),
            'test': False,
            'train_matrix_uuid': matrix_uuid, 
            'training_label_timespan': self.training_label_timespan,
            'model_comment': retrain_model_comment,
        }
        
        # get the random seed from the last split 
        last_split_train_matrix_uuid, last_split_matrix_metadata = train_matrix_info_from_model_id(
            self.db_engine, 
            model_id=self.model_group_info['model_id_last_split']
        )

        random_seed = self.model_trainer.get_or_generate_random_seed( 
            model_group_id=self.model_group_id, 
            matrix_metadata=last_split_matrix_metadata, 
            train_matrix_uuid=last_split_train_matrix_uuid
        )
        
        # create retrain model hash
        retrain_model_hash = self.model_trainer._model_hash(
                self.matrix_storage_engine.get_store(matrix_uuid).metadata,
                class_path=self.model_group_info['model_type'],
                parameters=self.model_group_info['hyperparameters'],
                random_seed=random_seed,
            ) 

        associate_models_with_retrain(self.retrain_hash, (retrain_model_hash, ), self.db_engine)

        record_model_building_started(run_id, self.db_engine)
        retrain_model_id = self.model_trainer.process_train_task(
            matrix_store=self.matrix_storage_engine.get_store(matrix_uuid), 
            class_path=self.model_group_info['model_type'], 
            parameters=self.model_group_info['hyperparameters'], 
            model_hash=retrain_model_hash,
            misc_db_parameters=misc_db_parameters, 
            random_seed=random_seed, 
            retrain=True,
            model_group_id=self.model_group_id
        )

        self.retrain_model_hash = retrieve_model_hash_from_id(self.db_engine, retrain_model_id)
        self.retrain_matrix_uuid = matrix_uuid
        self.retrain_model_id = retrain_model_id
        return {'retrain_model_comment': retrain_model_comment, 'retrain_model_id': retrain_model_id}

    def predict(self, prediction_date):
        """Predict forward by creating a matrix using as_of_date = prediction_date and applying the retrain model on it

        Args:
            prediction_date(str)
        """
        cohort_table_name = f"triage_production.cohort_{self.cohort_name}_predict"

        # 1. Generate cohort
        self.generate_entity_date_table(prediction_date, cohort_table_name)

        # 2. Generate feature aggregations
        collate_aggregations = self.get_collate_aggregations(prediction_date, cohort_table_name)
        self.feature_generator.process_table_tasks(
            self.feature_generator.generate_all_table_tasks(
                collate_aggregations,
                task_type='aggregation'
            )
        )
        # 3. Reconstruct feature disctionary from feature_names and generate imputation
        reconstructed_feature_dict, imputation_table_tasks = self.get_feature_dict_and_imputation_task(
                collate_aggregations, 
                self.retrain_model_id
        )
        self.feature_generator.process_table_tasks(imputation_table_tasks)
 
        # 4. Build matrix
        db_config = {
            "features_schema_name": "triage_production",
            "labels_schema_name": "public",
            "cohort_table_name": cohort_table_name,
        }

        matrix_builder = MatrixBuilder(
            db_config=db_config,
            matrix_storage_engine=self.matrix_storage_engine,
            engine=self.db_engine,
            experiment_hash=None,
            replace=True,
        )
        # Use timechop to get the time definition for production
        temporal_config = self.get_temporal_config_for_retrain(dt_from_str(prediction_date))
        timechopper = Timechop(**temporal_config)

        retrain_config = get_retrain_config_from_model_id(self.db_engine, self.retrain_model_id)

        prod_definitions = timechopper.define_test_matrices(
            train_test_split_time=dt_from_str(prediction_date), 
            test_duration=retrain_config['test_duration'],
            test_label_timespan=retrain_config['test_label_timespan']
        )
        last_split_definition = prod_definitions[-1]
        matrix_metadata = Planner.make_metadata(
            matrix_definition=last_split_definition,
            feature_dictionary=reconstructed_feature_dict,
            label_name=self.label_name,
            label_type='binary',
            cohort_name=self.cohort_name,
            matrix_type='production',
            feature_start_time=self.feature_start_time,
            user_metadata=self.user_metadata,
        )
    
        matrix_metadata['matrix_id'] = str(prediction_date) +  f'_model_id_{self.retrain_model_id}' + '_risklist'

        matrix_uuid = filename_friendly_hash(matrix_metadata)
    
        matrix_builder.build_matrix(
            as_of_times=[prediction_date],
            label_name=self.label_name,
            label_type='binary',
            feature_dictionary=reconstructed_feature_dict,
            matrix_metadata=matrix_metadata,
            matrix_uuid=matrix_uuid,
            matrix_type="production",
        )
        
        # 5. Predict the risk score for production
        predictor = Predictor(
            model_storage_engine=self.project_storage.model_storage_engine(),
            db_engine=self.db_engine,
            rank_order='best'
        )

        predictor.predict(
            model_id=self.retrain_model_id,
            matrix_store=self.matrix_storage_engine.get_store(matrix_uuid),
            misc_db_parameters={},
            train_matrix_columns=self.matrix_storage_engine.get_store(self.retrain_matrix_uuid).columns(),
        )
        self.predict_matrix_uuid = matrix_uuid



from __future__ import with_statement

import os

import yaml
import re
from alembic import context
from sqlalchemy import create_engine
from sqlalchemy import pool
from sqlalchemy.engine.url import URL

from triage.component.results_schema import Base


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

def get_excludes_from_config(config_, type_="tables"):
    excludes = config_.get(type_, None)
    if excludes is not None:
        excludes = excludes.split(",")
    excludes = excludes or []
    return excludes


excluded_tables = get_excludes_from_config(config.get_section('exclude'), "tables")
excluded_indices = get_excludes_from_config(config.get_section('exclude'), "indices")


def include_object(obj, name, type_, reflected, compare_to):
    if type_ == "table":
        for table_pat in excluded_tables:
            if re.match(table_pat, name):
                return False
        return True

    elif type_ == "index":
        for index_pat in excluded_indices:
            if re.match(index_pat, name):
                return False
        return True

    else:
        return True


url = None

if "url" in config.attributes:
    url = config.attributes["url"]

if not url:
    url = os.environ.get("DBURL", None)

if not url:
    db_config_file = context.get_x_argument("db_config_file").get(
        "db_config_file", None
    )
    if not db_config_file:
        raise ValueError("No database connection information found")

    with open(db_config_file) as fd:
        config = yaml.full_load(fd)
        url = URL(
            "postgres",
            host=config["host"],
            username=config["user"],
            database=config["db"],
            password=config["pass"],
            port=config["port"],
        )

def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        version_table="results_schema_versions",
        include_object=include_object,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    connectable = create_engine(url, poolclass=pool.NullPool)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table="results_schema_versions",
            include_schemas=True,
            include_object=include_object,
        )
        connection.execute('set search_path to "{}", public'.format("results"))

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import sys
import pandas as pd 
import logging
import sqlalchemy
import yaml
import argparse

from triage.component.catwalk.predictors import Predictor
from triage.component.catwalk.storage import ProjectStorage
from triage import create_engine


def _fetch_relevant_model_matrix_info(db_engine, model_groups, experiment_hashes=None):
    """ For the given experiment and model groups, fetch the model_ids, and match them with their train/test matrix pairs 
        Args:
            model_groups (List[int]): A list of model groups in the experiment
            experiment_hashes (List[str]): Optional. A list of experiment hashes we are interested in.
                If this is provided, only the model_ids that are relevant to the given experiments will be returned

        Return: 
            A DataFrame that contains the relevant model_ids, train_end_times, and matrix information
    """

    q = """
        SELECT 
            model_id,
            train_end_time,
            model_hash,
            model_group_id,
            train_matrix_uuid,
            c.matrix_uuid as test_matrix_uuid
        FROM triage_metadata.experiment_models a 
            JOIN triage_metadata.models b using(model_hash)
                JOIN test_results.prediction_metadata c using(model_id)
        WHERE model_group_id in ({model_groups})"""

    args_dict = {
        'model_groups': ', '.join([str(x) for x in model_groups])
    }

    if experiment_hashes is not None:
        q = q + " AND experiment_hash in ({experiment_hashes})"
        args_dict['experiment_hashes'] = ', '.join(["'" + str(x) + "'" for x in experiment_hashes])
    
    q = q.format(**args_dict)

    return pd.read_sql(q, db_engine)


def _summary_of_models(models_df):
    """generates logging summaries and warnings about the models being scored"""

    # how many models per model group?
    cts = models_df.groupby('model_group_id').count()['model_id']
    cts.name = 'num_models'
    model_counts = cts.unique()
    if len(model_counts) > 1:
        logging.warning('No. of models is different between model groups. See below: \n {}'.format(cts.reset_index()))
    else:
        logging.info('Each model group contains {} models'.format(model_counts[0]))

    # how many models per model group at each train_end_time?
    # Sometimes there are more than one model for each model group with different random seeds
    time_cts = models_df.groupby(['train_end_time', 'model_group_id']).count()['model_id']
    time_cts.name = 'num_models'
    msk = time_cts > 1

    if len(time_cts[msk]) > 0:
        logging.warning('There are model groups with more than one model per train_end_time. See below: \n {}'.format(time_cts[msk].reset_index()))
        
    
def add_predictions(db_engine, model_groups, project_path, experiment_hashes=None, train_end_times_range=None, rank_order='worst', replace=True):
    """ For a set of modl_groups generate test predictions and write to DB
        Args:
            db_engine: Sqlalchemy engine
            model_groups (list): The list of model group ids we are interested in (ideally, chosen through audition)
            project_path (str): Path where the created matrices and trained model objects are stored for the experiment
            experiment_hashes (List[str]): Optional. hash(es) of the experiments we are interested in. Can be used to narrow down the model_ids in the model groups specified
            range_train_end_times (Dict): Optional. If provided, only the models with train_end_times that fall in the range are scored. 
                                        This too, helps narrow down model_ids in the model groups specified.
                                        A dictionary with two possible keys 'range_start_date' and 'range_end_date'. Either or both could be set
            rank_order (str) : How to deal with ties in the scores. 
            replace (bool) : Whether to overwrite the preditctions for a model_id, if already found in the DB.

        Returns: None
            This directly writes to the test_results.predictions table
    """

    model_matrix_info = _fetch_relevant_model_matrix_info(
        db_engine=db_engine,
        model_groups=model_groups,
        experiment_hashes=experiment_hashes
    )

    # If we are only generating predictions for a specific time range
    if train_end_times_range is not None: 
        if 'range_start_date' in train_end_times_range:
            range_start = train_end_times_range['range_start_date']
            msk = (model_matrix_info['train_end_time'] >= range_start)
            logging.info('Filtering out models with a train_end_time before {}'.format(range_start))

            model_matrix_info = model_matrix_info[msk]

        if 'range_end_date' in train_end_times_range:       
            range_end = train_end_times_range['range_end_date']
            msk = (model_matrix_info['train_end_time'] <= range_end)
            logging.info('Filtering out models with a train_end_time after {}'.format(range_end))

            model_matrix_info = model_matrix_info[msk]

    if len(model_matrix_info)==0:
        raise ValueError('Configis not valid. No models were found!')
    
    # Al the model groups specified in the config file should valid (even if the experiment_hashes and train_end_times are specified)
    not_fetched_model_grps = [x for x in model_groups if not x in model_matrix_info['model_group_id'].unique()]
    
    if len(not_fetched_model_grps) > 0:
        raise ValueError('The config is not valid. No models were found for the model group(s) {}. All specified model groups should be present'.format(not_fetched_model_grps))
    
    logging.info('Scoring {} model ids'.format(len(model_matrix_info)))

    # summary of the models that we are scoring. To check any special things worth noting
    _summary_of_models(model_matrix_info)
    
    logging.info('Instantiating storage engines and the predictor')
    
    # Storage objects to handle already stored models and matrices
    project_storage = ProjectStorage(project_path)
    model_storage_engine = project_storage.model_storage_engine()
    matrix_storage_engine = project_storage.matrix_storage_engine()

    # Prediction generation is handled by the Predictor class in catwalk
    predictor = Predictor(
        model_storage_engine=model_storage_engine,
        db_engine=db_engine,
        rank_order=rank_order,
        replace=replace,
        save_predictions=True
    )

    # Organizing prediction run over unique (train_mat, test_mat) pairs
    # This is to reduce no. the times the matrices get loaded to memory
    groupby_obj = model_matrix_info.groupby(['train_matrix_uuid', 'test_matrix_uuid'])

    for group, _ in groupby_obj:
        train_uuid = group[0]
        test_uuid = group[1]

        df_grp = groupby_obj.get_group(group)

        logging.info('Processing {} model_ids for train matrix {} and test matrix {}'.format(
            len(df_grp), train_uuid, test_uuid
        ))

        # train_matrix_store = matrix_storage_engine.get_store(matrix_uuid=train_uuid)

        # To ensure that the column order we use for predictions match the order we used in model training
        # train_matrix_columns = list(train_matrix_store.design_matrix.columns)
        
        test_matrix_store = matrix_storage_engine.get_store(matrix_uuid=test_uuid)
        train_matrix_columns = list(test_matrix_store.design_matrix.columns)

        for model_id in df_grp['model_id'].tolist():
            logging.info('Writing predictions for model_id {}'.format(model_id))
            predictor.predict(
                model_id=model_id,
                matrix_store=test_matrix_store,
                train_matrix_columns=train_matrix_columns,
                misc_db_parameters={}
            )

    logging.info('Successfully generated predictions for {} models!'.format(len(model_matrix_info)))



from datetime import datetime, timedelta

import pytest
import testing.postgresql
from sqlalchemy.engine import create_engine

from triage.component.architect.entity_date_table_generators import EntityDateTableGenerator, SubsetEntityDateTableGenerator

from . import utils


def test_empty_output():
    """An empty cohort table eagerly produces an error.

    (Rather than allowing execution to proceed.)

    """
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", [])
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_cohort",
        )

        with pytest.raises(ValueError):
            # Request time outside of available intervals
            table_generator.generate_entity_date_table([datetime(2015, 12, 31)])

        (cohort_count,) = engine.execute(
            f"""\
            select count(*) from {table_generator.entity_date_table_name}
        """
        ).first()

        assert cohort_count == 0

        engine.dispose()


def test_entity_date_table_generator_replace():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_entity_date",
            replace=True
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output
        utils.assert_index(engine, table_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, table_generator.entity_date_table_name, "as_of_date")

        table_generator.generate_entity_date_table(as_of_dates)
        assert results == expected_output


def test_entity_date_table_generator_noreplace():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_entity_date",
            replace=False
        )

        # 1. generate a cohort for a subset of as-of-dates
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output
        utils.assert_index(engine, table_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, table_generator.entity_date_table_name, "as_of_date")

        table_generator.generate_entity_date_table(as_of_dates)
        assert results == expected_output

        # 2. generate a cohort for a different subset of as-of-dates,
        # actually including an overlap to make sure that it doesn't double-insert anything
        as_of_dates = [
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output


def test_entity_date_table_generator_from_labels():
    labels_data = [
        (1, datetime(2016, 1, 1), timedelta(180), 'outcome', 'binary', 0),
        (1, datetime(2016, 4, 1), timedelta(180), 'outcome', 'binary', 1),
        (1, datetime(2016, 3, 1), timedelta(180), 'outcome', 'binary', 0),
        (2, datetime(2016, 1, 1), timedelta(180), 'outcome', 'binary', 0),
        (2, datetime(2016, 1, 1), timedelta(180), 'outcome', 'binary', 1),
        (3, datetime(2016, 1, 1), timedelta(180), 'outcome', 'binary', 0),
        (5, datetime(2016, 3, 1), timedelta(180), 'outcome', 'binary', 0),
        (5, datetime(2016, 4, 1), timedelta(180), 'outcome', 'binary', 1),
        (1, datetime(2016, 1, 1), timedelta(90), 'outcome', 'binary', 0),
        (1, datetime(2016, 4, 1), timedelta(90), 'outcome', 'binary', 0),
        (1, datetime(2016, 3, 1), timedelta(90), 'outcome', 'binary', 1),
        (2, datetime(2016, 1, 1), timedelta(90), 'outcome', 'binary', 0),
        (2, datetime(2016, 1, 1), timedelta(90), 'outcome', 'binary', 1),
        (3, datetime(2016, 1, 1), timedelta(90), 'outcome', 'binary', 0),
        (5, datetime(2016, 3, 1), timedelta(90), 'outcome', 'binary', 0),
        (5, datetime(2016, 4, 1), timedelta(90), 'outcome', 'binary', 0),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        labels_table_name = utils.create_labels(engine, labels_data)
        table_generator = EntityDateTableGenerator(
            query=None,
            labels_table_name=labels_table_name,
            db_engine=engine,
            entity_date_table_name="exp_hash_entity_date",
            replace=False
        )
        table_generator.generate_entity_date_table([])
        expected_output = [
            (1, datetime(2016, 1, 1)),
            (1, datetime(2016, 3, 1)),
            (1, datetime(2016, 4, 1)),
            (2, datetime(2016, 1, 1)),
            (3, datetime(2016, 1, 1)),
            (5, datetime(2016, 3, 1)),
            (5, datetime(2016, 4, 1)),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output


def test_subset_evens_cohort():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 3, 1), True),
        (3, datetime(2016, 1, 1), True),
        (4, datetime(2016, 3, 1), True),
        (4, datetime(2016, 4, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)

        # create cohort table 
        engine.execute(
        """
            create table cohort (
                entity_id int,
                as_of_date date,
                active bool
            )
        """
        )
        for row in input_data:
            engine.execute("insert into cohort values (%s, %s, %s)", row)
        
        as_of_dates = [

            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
        ]
        ### test evens
        subset_generator = SubsetEntityDateTableGenerator(
            query="select distinct entity_id, outcome_date from events where entity_id %% 2 = 0 and outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_subset_entity_date",
            replace=True,
            cohort_table="cohort")
        subset_generator.generate_entity_date_table(as_of_dates)
        
        expected_output = [
            (2, datetime(2016, 3, 1), True),
            (4, datetime(2016, 4, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {subset_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )

        assert results == expected_output 
        utils.assert_index(engine, subset_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, subset_generator.entity_date_table_name, "as_of_date")
        


def test_subset_odds_cohort():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 3, 1), True),
        (3, datetime(2016, 3, 1), True),
        (4, datetime(2016, 3, 1), True),
        (4, datetime(2016, 4, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)

        # create cohort table 
        engine.execute(
        """
            create table cohort (
                entity_id int,
                as_of_date date,
                active bool
            )
        """
        )
        for row in input_data:
            engine.execute("insert into cohort values (%s, %s, %s)", row)
        
        as_of_dates = [
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
        ]
        ### test odds
        subset_generator = SubsetEntityDateTableGenerator(
            query="select distinct entity_id, outcome_date from events where entity_id %% 2 = 1 and outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_subset_entity_date",
            replace=True,
            cohort_table="cohort")
        subset_generator.generate_entity_date_table(as_of_dates)
        
        expected_output = [
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (5, datetime(2016, 4, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {subset_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )

        assert results == expected_output 
        utils.assert_index(engine, subset_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, subset_generator.entity_date_table_name, "as_of_date")


def test_subset_empty():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 3, 1), True),
        (3, datetime(2016, 3, 1), True),
        (4, datetime(2016, 3, 1), True),
        (4, datetime(2016, 4, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)

        # create cohort table 
        engine.execute(
        """
            create table cohort (
                entity_id int,
                as_of_date date,
                active bool
            )
        """
        )
        for row in input_data:
            engine.execute("insert into cohort values (%s, %s, %s)", row)
        
        as_of_dates = [
            datetime(2016, 1, 1),
        ]
        ### test evens
        subset_generator = SubsetEntityDateTableGenerator(
            query="select distinct entity_id, outcome_date from events where entity_id %% 2 = 1 and outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_subset_entity_date",
            replace=True,
            cohort_table="cohort")
        subset_generator.generate_entity_date_table(as_of_dates)
        
        expected_output = []
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {subset_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )

        assert results == expected_output 
        utils.assert_index(engine, subset_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, subset_generator.entity_date_table_name, "as_of_date")


def test_subset_all_entities_in_cohort():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 3, 1), True),
        (3, datetime(2016, 3, 1), True),
        (4, datetime(2016, 3, 1), True),
        (4, datetime(2016, 4, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)

        # create cohort table 
        engine.execute(
        """
            create table cohort (
                entity_id int,
                as_of_date date,
                active bool
            )
        """
        )
        for row in input_data:
            engine.execute("insert into cohort values (%s, %s, %s)", row)
        
        as_of_dates = [
            datetime(2016, 1, 1),
        ]
        ### test evens
        subset_generator = SubsetEntityDateTableGenerator(
            query="select distinct entity_id, outcome_date from events where entity_id %% 2 = 0 and outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_subset_entity_date",
            replace=True,
            cohort_table="cohort")
        subset_generator.generate_entity_date_table(as_of_dates)
        
        expected_output = set([2, 4])
        results = set(list(
            engine.execute(
                f"""
                select distinct entity_id from {subset_generator.entity_date_table_name}
            """
            )
        ))

        assert results.issubset(expected_output) == True


"""empty message

Revision ID: b097e47ba829
Revises: 45219f25072b
Create Date: 2021-05-20 15:40:47.288721

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b097e47ba829'
down_revision = '45219f25072b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('experiment_runs', 'matrices_errored',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'matrices_made',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'matrices_skipped',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_errored',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_made',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_skipped',
               existing_type=sa.INTEGER(),
               nullable=True,
               schema='triage_metadata')
    op.drop_column('experiments', 'random_seed', schema='triage_metadata')
    op.create_index(op.f('ix_triage_metadata_matrices_matrix_uuid'), 'matrices', ['matrix_uuid'], unique=True, schema='triage_metadata')
    op.drop_index('ix_model_metadata_matrices_matrix_uuid', table_name='matrices', schema='triage_metadata')
    op.create_index(op.f('ix_triage_metadata_models_model_hash'), 'models', ['model_hash'], unique=True, schema='triage_metadata')
    op.drop_index('ix_model_metadata_models_model_hash', table_name='models', schema='triage_metadata')
    op.create_foreign_key('models_built_in_experiment_run_fkey', 'models', 'experiment_runs', ['built_in_experiment_run'], ['id'], source_schema='triage_metadata', referent_schema='triage_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('models_built_in_experiment_run_fkey', 'models', schema='triage_metadata', type_='foreignkey')
    op.create_index('ix_model_metadata_models_model_hash', 'models', ['model_hash'], unique=True, schema='triage_metadata')
    op.drop_index(op.f('ix_triage_metadata_models_model_hash'), table_name='models', schema='triage_metadata')
    op.create_index('ix_model_metadata_matrices_matrix_uuid', 'matrices', ['matrix_uuid'], unique=True, schema='triage_metadata')
    op.drop_index(op.f('ix_triage_metadata_matrices_matrix_uuid'), table_name='matrices', schema='triage_metadata')
    op.add_column('experiments', sa.Column('random_seed', sa.INTEGER(), autoincrement=False, nullable=True), schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_skipped',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_made',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'models_errored',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'matrices_skipped',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'matrices_made',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    op.alter_column('experiment_runs', 'matrices_errored',
               existing_type=sa.INTEGER(),
               nullable=False,
               schema='triage_metadata')
    # ### end Alembic commands ###



"""Add production prediction metadata

Revision ID: 670289044eb2
Revises: ce5b50ffa8e2
Create Date: 2021-01-08 22:27:23.433813

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '670289044eb2'
down_revision = 'ce5b50ffa8e2'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('prediction_metadata',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('matrix_uuid', sa.Text(), nullable=False),
    sa.Column('tiebreaker_ordering', sa.Text(), nullable=True),
    sa.Column('random_seed', sa.Integer(), nullable=True),
    sa.Column('predictions_saved', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['triage_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['triage_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'matrix_uuid'),
    schema='production'
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('prediction_metadata', schema='production')
    # ### end Alembic commands ###



import matplotlib
import matplotlib.dates as md
import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objects as go
matplotlib.use('Agg')

from datetime import datetime
from plotly.subplots import make_subplots
from triage.util.conf import convert_str_to_relativedelta


FIG_SIZE = (32, 16)


def visualize_chops(chopper, show_as_of_times=True, show_boundaries=True, save_target=None):
    """Visualize time chops of a given Timechop object using matplotlib

    Args:
        chopper (triage.component.timechop.Timechop): A fully-configured Timechop object
        show_as_of_times (bool): Whether or not to draw horizontal lines
            for as-of-times
        show_boundaries (bool): Whether or not to show a rectangle around matrices
            and dashed lines around feature/label boundaries
        save_target (str or file-like object): A save target for matplotlib to save
            the figure to. Defaults to None, which won't save anything
    """
    chops = chopper.chop_time()

    chops.reverse()

    fig, ax = plt.subplots(nrows=len(chops), sharex=True, sharey=True, squeeze=False, figsize=FIG_SIZE)

    for idx, chop in enumerate(chops):
        train_as_of_times = chop["train_matrix"]["as_of_times"]
        test_as_of_times = chop["test_matrices"][0]["as_of_times"]

        test_label_timespan = chop["test_matrices"][0]["test_label_timespan"]
        training_label_timespan = chop["train_matrix"]["training_label_timespan"]

        color_rgb = np.random.random(3)

        if show_as_of_times:
            # Train matrix (as_of_times)
            ax[idx][0].hlines(
                [x for x in range(len(train_as_of_times))],
                [x.date() for x in train_as_of_times],
                [
                    x.date() + convert_str_to_relativedelta(training_label_timespan)
                    for x in train_as_of_times
                ],
                linewidth=3,
                color=color_rgb,
                label=f"train_{idx}",
            )

            # Test matrix
            ax[idx][0].hlines(
                [x for x in range(len(test_as_of_times))],
                [x.date() for x in test_as_of_times],
                [
                    x.date() + convert_str_to_relativedelta(test_label_timespan)
                    for x in test_as_of_times
                ],
                linewidth=3,
                color=color_rgb,
                label=f"test_{idx}",
            )

        if show_boundaries:
            # Limits: train
            ax[idx][0].axvspan(
                chop["train_matrix"]["first_as_of_time"],
                chop["train_matrix"]["last_as_of_time"],
                color=color_rgb,
                alpha=0.3,
            )

            ax[idx][0].axvline(
                chop["train_matrix"]["matrix_info_end_time"], color="k", linestyle="--"
            )

            # Limits: test
            ax[idx][0].axvspan(
                chop["test_matrices"][0]["first_as_of_time"],
                chop["test_matrices"][0]["last_as_of_time"],
                color=color_rgb,
                alpha=0.3,
            )

            ax[idx][0].axvline(
                chop["feature_start_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["feature_end_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["label_start_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["label_end_time"], color="k", linestyle="--", alpha=0.2
            )

            ax[idx][0].axvline(
                chop["test_matrices"][0]["matrix_info_end_time"],
                color="k",
                linestyle="--",
            )

        ax[idx][0].yaxis.set_major_locator(plt.NullLocator())
        ax[idx][0].yaxis.set_label_position("right")
        ax[idx][0].set_ylabel(f'Label timespan \n {test_label_timespan} (test), {training_label_timespan} (training)',
                              rotation="vertical", labelpad=30)

        ax[idx][0].xaxis.set_major_formatter(md.DateFormatter("%Y"))
        ax[idx][0].xaxis.set_major_locator(md.YearLocator())
        ax[idx][0].xaxis.set_minor_locator(md.MonthLocator())

    ax[0][0].set_title("Timechop: Temporal cross-validation blocks")
    fig.subplots_adjust(hspace=0)
    plt.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
    if save_target:
        plt.savefig(save_target)
    plt.show()


def visualize_chops_plotly(chopper, selected_splits=None, show_label_timespans=True, show_boxes=True, show_annotations=True):
    """Visualize time chops of a given Timechop object using plotly, to get an interactive output
    Args:
        chopper (triage.component.timechop.Timechop): A fully-configured Timechop object
        selected_splits (list): Indices of train-val sets to plot. E.g. [0, 1, 2] plots the 3 most recent splits, [0,-1] plots the first and last splits.
            Defaults to None, which plots all splits.
        show_label_timespans (bool): Whether or not to draw horizontal lines to show label timespan
            for as-of-times
        show_boxes (bool): Whether or not to show a rectangle highlighting train-test matrices
        show_annotations (bool): Whether or not to add annotations on the latest split, showing what each of the timechop parameters mean
    """
    chops = chopper.chop_time()
    chops.reverse() # reverse to get the most recent set first

    # Subset to relevant splits if arg specified, and generate titles for each split
    if selected_splits is not None:
      chops = [chops[i] for i in selected_splits]
      titles = tuple(f"Train-Validation Split {i+1}" for i in selected_splits)
    else:
      titles = tuple(f"Train-Validation Split {i+1}" for i in range(len(chops)))

    fig = make_subplots(rows=len(chops), 
                        cols=1,
                        shared_xaxes=True,
                        shared_yaxes=True,
                        # vertical_spacing=0.05,
                        subplot_titles=titles
        ) # adds titles for each subplot

    # For each train-val split
    for idx, chop in enumerate(chops):
        train_as_of_times = chop["train_matrix"]["as_of_times"]
        test_as_of_times = chop["test_matrices"][0]["as_of_times"]

        test_label_timespan = chop["test_matrices"][0]["test_label_timespan"]
        training_label_timespan = chop["train_matrix"]["training_label_timespan"]

        # Colors for train/test 
        train_color = "rgba(3, 37, 126" # dark blue (left open because we add an opacity argument below)
        test_color = "rgba(139, 0, 0" # magenta (left open because we add an opacity argument below)
        as_of_date_marker_opacity = ', 1)' # the extra ', 1)' defines opacity. 100% solid for markers
        label_line_opacity = ', 0.3)' # 30% opacity for the label lines
        rectangle_fill_opacity = ', 0.15)' # 15% opacity for rectangle fill

        train_as_of_date_color = train_color + as_of_date_marker_opacity
        train_label_period_color = train_color + label_line_opacity
        train_rectangle_fill = train_color + rectangle_fill_opacity
        test_as_of_date_color = test_color + as_of_date_marker_opacity
        test_label_period_color = test_color + label_line_opacity
        test_rectangle_fill = test_color + rectangle_fill_opacity

        # Show legend only if idx = 0 (i.e. first train-val set we are displaying)
        if idx == 0:
          # Train set as-of-date markers
          fig.add_trace(
              go.Scatter(x=[x.date() for x in train_as_of_times], 
                        y=[x for x in range(len(train_as_of_times))], 
                        mode='markers',
                        marker=dict(color=train_as_of_date_color),
                        name='Training as-of-date',
                        showlegend=True,
                        hovertemplate="%{x}<extra></extra>" # the extra extra tag gets rid of a default 'trace' line in the hover output and just shows 'x', the date
                        ), 
              row=idx+1, # row and column of the subplots to add this trace object to
              col=1
              )
          # Validation set as-of-date markers
          fig.add_trace(
            go.Scatter(x=[x for x in test_as_of_times], 
                      y=[x for x in range(len(test_as_of_times))], 
                      mode='markers',
                      name='Validation as-of-date',
                      showlegend=True,
                      marker=dict(color=test_as_of_date_color),
                      hovertemplate="%{x}<extra></extra>"),
            row=idx+1,
            col=1
            )
        # Suppress legend if not the first subplot; only difference with above is showlegend=False (note, anytime we add a trace, we have to set showlegend=False to suppress useless info in the legend)
        else:
          # Train set as-of-date markers
          fig.add_trace(
              go.Scatter(x=[x.date() for x in train_as_of_times], 
                        y=[x for x in range(len(train_as_of_times))], 
                        mode='markers',
                        marker=dict(color=train_as_of_date_color),
                        name='Training as-of-date',
                        showlegend=False,
                        hovertemplate="%{x}<extra></extra>" # the extra extra tag gets rid of a default 'trace' line in the hover output and just shows 'x', the date
                        ), 
              row=idx+1, # row and column of the subplots to add this trace object to
              col=1
              )

          # Validation set as-of-date markers
          fig.add_trace(
            go.Scatter(x=[x for x in test_as_of_times], 
                      y=[x for x in range(len(test_as_of_times))], 
                      mode='markers',
                      name='Validation as-of-date',
                      showlegend=False,
                      marker=dict(color=test_as_of_date_color),
                      hovertemplate="%{x}<extra></extra>"),
            row=idx+1,
            col=1
            )


        # Add test_durations annotation if option selected
        if idx == 0 and show_annotations==True:

          # Add a dashed line to show test_durations span
          x0 = test_as_of_times[0]
          x1 = test_as_of_times[-1]
          x_mid = x0 + (x1-x0)/2
          y = -1 # place the test durations labeling below the graph
          fig.add_shape(type='line', x0=x0, x1=x1, y0=y, y1=y, line={'color': 'green'}, row=idx+1, col=1)
          fig.add_annotation(x=x_mid, y=y-1, text=f"Test duration: {chop['test_matrices'][0]['test_duration']}", showarrow=False)

        # Add label timespan lines if option selected
        if show_label_timespans is True:

          # For training as_of_dates
          for i in range(len(train_as_of_times)):
            fig.add_trace(
                go.Scatter(
                    x=[train_as_of_times[i].date(), train_as_of_times[i].date() + convert_str_to_relativedelta(training_label_timespan)],
                    y=[i,i],
                    marker=dict(color=train_label_period_color, line=dict(color=train_label_period_color)),
                    hovertemplate="%{x}<extra></extra>",
                    showlegend=False
                ),
              row=idx+1,
              col=1 
            )

            # Add annotation showing train label timespan on first bar in first train-val set (if option specified) 
            if i == len(train_as_of_times)-1 and idx == 0 and show_annotations==True:

              # Have the x in between the label timespan
              x0 = train_as_of_times[i].date()
              x1 = train_as_of_times[i].date() + convert_str_to_relativedelta(training_label_timespan)
              x_pos = x0 + (x1 - x0)/2

              # Position at a y-value above the bar
              y_pos = i
              fig.add_annotation(x=x_pos, y=y_pos, text='Label timespan', showarrow=True, arrowhead=1, row=idx+1, col=1)

          # For test as_of_dates
          for i in range(len(test_as_of_times)):
            fig.add_trace(
                go.Scatter(
                    x=[test_as_of_times[i].date(), test_as_of_times[i].date() + convert_str_to_relativedelta(test_label_timespan)],
                    y=[i,i],
                    marker=dict(color= test_label_period_color, line=dict(color= test_label_period_color)),
                    showlegend=False,
                    hovertemplate="%{x}<extra></extra>"),
              row=idx+1,
              col=1 
            )

            # Add annotation showing test label timespan on first bar in first train-val set (if option specified) 
            if i == len(test_as_of_times)-1 and idx == 0 and show_annotations==True:

                # Have the x in between the label timespan
                x0 = test_as_of_times[i].date()
                x1 = test_as_of_times[i].date() + convert_str_to_relativedelta(test_label_timespan)
                x_pos = x0 + (x1 - x0)/2

                # Position at a y-value above the bar
                y_pos = i
                fig.add_annotation(x=x_pos, y=y_pos, text='Label timespan', showarrow=True, arrowhead=1, row=idx+1, col=1)

        # Add rectangles/boxes to mark train-test matrices
        if show_boxes is True:

          # Training matrix rectangle
          # Rectangle params
          x0 = min(train_as_of_times).date()
          x1 = max(train_as_of_times).date() + convert_str_to_relativedelta(training_label_timespan)
          y = max(len(test_as_of_times), len(train_as_of_times))

          fig.add_trace(
              go.Scatter(x =[x0,x0,x1,x1,x0], y=[0,y,y,0,0], 
                        fill='toself', fillcolor=train_rectangle_fill,
                        showlegend=False,
                        marker=dict(color='rgba(0,255,0,0)', line=dict(color='rgba(0,255,0,0)')), # setting 0 opacity so we don't see the lines or markers
                        hoverinfo='skip'), 
              row=idx+1,
              col=1,
          )

          # #Add annotated text to the middle of the training set rectangle -> this code works, but the positioning is a bit weird, so need to tweak
          # middle_index = round(len(train_as_of_times)/2)
          # x_middle = train_as_of_times[middle_index].date() + convert_str_to_relativedelta(training_label_timespan)
          # fig.add_trace(
          #     go.Scatter(x =[x_middle], y=[y-1], 
          #               mode='text',
          #               text="Training Data",
          #               marker=dict(color='rgba(0,255,0,0)', line=dict(color='rgba(0,255,0,0)')), # setting 0 opacity so we don't see the lines 
          #               hoverinfo='skip'),
          #     row=idx+1,
          #     col=1,
          # )

          # Test set rectangle

          # Rectangle params
          x0 = min(test_as_of_times).date()
          x1 = max(test_as_of_times).date() + convert_str_to_relativedelta(test_label_timespan)
          y = max(len(test_as_of_times), len(train_as_of_times))

          fig.add_trace(
              go.Scatter(x =[x0,x0,x1,x1,x0], y=[0,y,y,0,0], 
                        fill='toself', fillcolor=test_rectangle_fill,
                        showlegend=False,
                        marker=dict(color='rgba(0,255,0,0)', line=dict(color='rgba(0,255,0,0)')), # setting 0 opacity so we don't see the lines 
                        hoverinfo='skip'),
              row=idx+1,
              col=1,
          )

          # #Add annotated text to the test set rectangle
          # middle_index = round(len(test_as_of_times)/2)
          # x_middle = test_as_of_times[middle_index].date() + convert_str_to_relativedelta(test_label_timespan)
          # fig.add_trace(
          #     go.Scatter(x =[x_middle], y=[y-1],  
          #               mode='text',
          #               text="Test Data",
          #               marker=dict(color='rgba(0,255,0,0)', line=dict(color='rgba(0,255,0,0)')), # setting 0 opacity so we don't see the lines 
          #               hoverinfo='skip'),
          #     row=idx+1,
          #     col=1,
          # )

    fig.update_layout(height=150*len(chops), width=800, showlegend=True)
    fig.show()




""" This is a module for moving the ad-hoc code we wrote in generating the modeling report"""
import verboselogs, logging

logger = verboselogs.VerboseLogger(__name__)

import pandas as pd
import json
import matplotlib.pyplot as plt
import seaborn as sns
import warnings

from matplotlib.lines import Line2D

from triage.component.timechop.plotting import visualize_chops_plotly
from triage.component.timechop import Timechop
from triage.component.audition.plotting import plot_cats

from triage.component.postmodeling.base import SingleModelAnalyzer
# from triage.component.postmodeling.postmodeling_analyzer import PostmodelingAnalyzer

model_name_abbrev = {
    'DummyClassifier': 'Dummy',
    'RandomForestClassifier': 'RF',
    'DecisionTreeClassifier': 'DT',
    'BaselineRankMultiFeature': 'Ranker',
    'ScaledLogisticRegression': 'LR',
    'XGBClassifier': 'XGB',
    'LightGBMClassifier':'LGBM'
}

def _format_model_name(long_name, model_group_id):
    
    # Get the child name fromt he import path
    child_name = long_name.split('.')[-1] 
    
    if child_name in model_name_abbrev:
        return str(model_group_id) + ': ' + model_name_abbrev[child_name]
    else:
        return str(model_group_id) + ': ' + child_name


def list_all_experiments(engine):
    """Generates a dataframe that lists the 
        experiment hash, last experiment run time, model comment
        for all experiments that have completed so far

    Args:
        engine (SQLAlchemy): Database engine
    """
    
    q = '''
        select 
            id,
            run_hash as experiment_hash,
            max(start_time) as most_recent_start_time,
            max(e.config ->> 'model_comment') as model_comment
        from triage_metadata.triage_runs tr join triage_metadata.experiments e on
        tr.run_hash = e.experiment_hash
        where current_status = 'completed'
        group by 1 order by 1 desc
    '''
    
    return pd.read_sql(q, engine)

def load_config(engine, experiment_hash):
    "return the experiment config"
    
    q = f'''
        select 
        config 
        from triage_metadata.experiments
        where experiment_hash = '{experiment_hash}'
    '''
    
    return pd.read_sql(q, engine).config.at[0]
    

def load_report_parameters_from_config(engine, experiment_hash):
    config = load_config(engine, experiment_hash=experiment_hash)
    
    d = dict()
    d['performance_metric'] = config['scoring'].get('priority_metric')
    d['threshold'] = config['scoring'].get('priority_parameter')
    
    bias_audit = config.get('bias_audit')
    d['bias_metric'] = None
    d['priority_groups'] = None 
    
    if bias_audit is not None:
        d['bias_metric'] = bias_audit.get('priority_metric')
        d['priority_groups'] = bias_audit.get('priority_groups')
    
    return d 
  
def get_most_recent_experiment_hash(engine):
    q = '''
        select 
            run_hash as experiment_hash
        from triage_metadata.triage_runs
        where current_status = 'completed'
        order by start_time desc       
    '''
    experiment_hash = pd.read_sql(q, engine)['experiment_hash'].iloc[0]
    logging.info(f'Using the experiment hash: {experiment_hash}')
    
    return experiment_hash







class ExperimentReport:
    
    def __init__(self, engine, 
                 experiment_hashes, 
                 performance_priority_metric='recall@', 
                 threshold="1_pct", 
                 bias_priority_metric='tpr_disparity', 
                 bias_priority_groups=None):
        self.engine = engine
        self.experiment_hashes = experiment_hashes
        
        # TODO - consider expanding to multiple metrics and/or thresholds
        self.performance_metric = performance_priority_metric
        self.threshold = threshold
        self.bias_metric = bias_priority_metric
        self.bias_groups = bias_priority_groups
    
    def timesplits(self):
        """Generate an interactive plot of the time splits used for cross validation

        Args:
            engine (SQLAlchemy): DB engine
            experiment_hash (str): Experiment hash we are interested in  
        """
        
        q = f'''
            select 
                config
            from triage_metadata.experiments where experiment_hash = '{self.experiment_hashes[0]}'
        '''

        experiment_config = pd.read_sql(q, self.engine).at[0, 'config']
        
        
        chops = Timechop(**experiment_config['temporal_config'])
        splits = len(chops.chop_time())

        print(f'There are {splits} train-validation splits')
        visualize_chops_plotly(
            chops
        )
           
    def cohorts(self, generate_plots=True):
        """Generate a summary of cohorts (size, baserate)

        Args:
            engine (SQLAlchemy): Database engine
            experiment_hash (str): a list of experiment hashes
        """
        
        get_labels_table_query = f"""
        select distinct labels_table_name from triage_metadata.triage_runs 
        where run_hash = '{self.experiment_hashes[0]}'
        """

        labels_table = pd.read_sql(get_labels_table_query, self.engine).labels_table_name.iloc[0]
        # print(labels_table)
        cohort_query = f"""
            select 
            label_name, 
            label_timespan, 
            as_of_date, 
            count(distinct entity_id) as cohort_size, 
            avg(label) as baserate 
            from public.{labels_table}
            group by 1,2,3 order by 1,2,3
        """

        df = pd.read_sql(cohort_query, self.engine)
        
        if generate_plots:
            fig, ax1 = plt.subplots(figsize=(7, 3), dpi=100)

            color='darkblue'
            sns.lineplot(
                data=df,
                x='as_of_date',
                y='cohort_size',
                ax=ax1,
                label='',
                color=color
            )

            ax1.axhline(y=df.cohort_size.mean(), color=color, alpha=0.4, linestyle='--', label='mean cohort size')
            ax1.set_title('Cohort Size and Baserate Over Time')
            ax1.set_xlabel('Time')
            ax1.set_ylabel('Cohort Size', color=color)
            ax1.tick_params(axis='y', labelcolor=color)

            ax2 = ax1.twinx()
            color='firebrick'

            sns.lineplot(
                data=df,
                x='as_of_date',
                y='baserate',
                ax=ax2,
                color=color
            )
            ax2.set_ylabel('Baserate (%)', color=color)
            ax2.tick_params(axis='y', labelcolor=color)
            ax2.axhline(y=df.baserate.mean(), color=color, alpha=0.4, linestyle='--')
                    
        return df

    def subsets(self):
        q = f'''
            select 
                distinct s.subset_hash, s.config, s.config ->> 'name' as subset_name,
                s.config ->> 'name'::text||'_'||s.subset_hash as table_name
            from triage_metadata.experiment_models join triage_metadata.models using(model_hash)
                left join test_results.evaluations e using(model_id)
                    inner join triage_metadata.subsets s on e.subset_hash = s.subset_hash 
            where experiment_hash in ('{"','".join(self.experiment_hashes)}')
        '''
        
        df = pd.read_sql(q, self.engine)
        
        df['table_name'] = df.table_name.apply(lambda x: f'subet_{x}')

        # TODO - add code to plot subset size relative to the cohort size
            
        return df.set_index('subset_hash')
        
    def model_groups(self):
        """Generate a summary of all the model groups (types and hyperparameters) built in the experiment

        Args:
            engine (SQLAlchemy): Database engine
            experiment_hashes (List[str]): List of experiment hashes
        """
                
        mg_query_selected_exp = ''' 
            select 
                model_group_id, 
                model_type, 
                hyperparameters,
                count(distinct model_id) as num_models, 
                count(distinct train_end_time) as num_time_splits
            from triage_metadata.experiment_models join triage_metadata.models using(model_hash)
            where experiment_hash in ('{}')
            group by 1, 2, 3
            order by 2
        '''

        model_groups = pd.read_sql(mg_query_selected_exp.format("','".join(self.experiment_hashes)), self.engine)
        
        #TODO - highlight model groups that don't have the "correct" number of models built. 
        # Some model objects can be missing, and some train_end_times can have multiple models
    
        return model_groups
    
    def models(self):
        """ List all of the models built in the experiment
        
        Args:
            engine (SQLAlchemy): Database engine
            experiment_hashes (List[str]): List of experiment hashes
        """
        
        q = f''' 
            select 
                model_id,
                model_group_id,
                model_type,
                hyperparameters,
                train_end_time
            from triage_metadata.experiment_models join triage_metadata.models using(model_hash)
            where experiment_hash in ('{"','".join(self.experiment_hashes)}')
            order by model_type
            '''

        df = pd.read_sql(q, self.engine)

        return df.sort_values(by=['model_group_id', 'train_end_time'])      
       
     
    def features(self):
        """ Generate a dataframe containing all the features that were built in an experimet
    
        Args:
            engine (SQLAlchemy): Database engine
            experiment_hash (str): Hash of the experiment 
                Limiting this to one experiment hash based on the assumption that if the feature 
                spaces are different the experiment hashes shouldn't be lumped together 
        """
        
        q = f'''
            select 
                config
            from triage_metadata.experiments where experiment_hash = '{self.experiment_hashes[0]}'
        '''

        experiment_config = pd.read_sql(q, self.engine).at[0, 'config']
        
        all_features = list()

        for fg in experiment_config['feature_aggregations']:
            
            for agg in fg.get('aggregates', {}):
                d = dict()
                d['feature_group'] = fg['prefix']
                if isinstance(agg['quantity'], dict):
                    d['feature_name'] = list(agg['quantity'].keys())[0]
                    
                else:
                    d['feature_name'] = agg['quantity']
            
                d['metrics'] = ', '.join(agg['metrics'])
                d['time_horizons'] = ', '.join(fg['intervals'])
                d['feature_type'] = 'continuous_aggregate'
                
                if agg.get('imputation') is None: 
                    imp = fg.get('aggregates_imputation')
                else:
                    imp = agg.get('imputation')
                d['imputation'] = json.dumps(imp)
                # d['num_predictors'] = len(agg['metrics']) * len(fg['intervals'])
                all_features.append(d)
                
            for cat in fg.get('categoricals', {}):
                d = dict()
                d['feature_group'] = fg['prefix']
                d['feature_name'] = cat['column']
                d['metrics'] = ', '.join(cat['metrics'])
                d['time_horizons'] = ', '.join(fg['intervals'])
                d['feature_type'] = 'categorical_aggregate'
                
                if agg.get('imputation') is None: 
                    imp = fg.get('categoricals_imputation')
                else:
                    imp = agg.get('imputation')
                d['imputation'] = json.dumps(imp)
                # d['num_predictors'] = 'number of categories'
                all_features.append(d)
                
        all_features = pd.DataFrame(all_features).sort_values('feature_group', ignore_index=True)
        
        return all_features

    def feature_missingness(self):
        """
        Generates the mean, min, max missingness degree for each feature. 
        Assumes that the current "features" schema holds the relevant features for the experiment
        """
        
        # TODO -- This code is not correct. This looks at all the feature tables rather than the config. 
        q = '''
        select 
        table_name 
        from information_schema.tables
        where table_schema = 'features'
        and table_name like '%%aggregation_imputed'
        '''

        feature_tables = pd.read_sql(q, self.engine)['table_name'].tolist()
        
        logging.info(f'Printing only features with missing values')
        
        column_names = dict()
        for table in feature_tables:
            q = f'''
                select 
                column_name
                FROM information_schema.columns
                WHERE table_schema = 'features'
                AND table_name   = '{table}'
                and (column_name like '%%_imp')   
            '''

            column_names[table] = pd.read_sql(q, self.engine).column_name.tolist()
            
        results = pd.DataFrame()

        for table_name, columns in column_names.items():
            # print(table_name)
            select_clause = '''
                select 
                as_of_date,
                count(distinct entity_id) as cohort_size
            '''
            imputation_counts = ''
            for col in columns:
                imputation_counts += f'''
                    ,(sum("{col}")::float / count(distinct entity_id)) * 100  as "{col[:-4]}"
                '''
            
            from_clause = f'''
                from features.{table_name}
                group by 1
            '''
            
            query = select_clause + imputation_counts + from_clause
            df = pd.read_sql(query, self.engine).set_index(['as_of_date', 'cohort_size'])

            if results.empty:
                results = df
            else:
                results = results.join(df)
                
        df = pd.concat([results.mean(), results.min(), results.max()], axis=1).fillna(0)
        df.columns = ['mean (%)', 'min (%)', 'max (%)']

        # df.style.applymap(lambda x: 'background-color : pink' if x>80 else '')
        df_ = df[df['mean (%)'] > 0]

        return df_.sort_values(by="mean (%)", ascending=False)
    
    
    def model_performance(self, metric=None, parameter=None, generate_plot=True):
        """ Generate an Audition type plot to display predictive performance over time for all model groups
        
        Args:
            metric (str): The metric we are intersted in (suppores 'precision@', 'recall@', 'auc_roc')
            parameter (str): The threshold, supports percentile ('_pct') and absolute ('_abs') thresholds
        """
        
        if metric is None:
            metric = self.performance_metric
        
        if parameter is None: 
            parameter = self.threshold
        
        # fetch model groups
        def fetch_evaluation_values(with_parameter=True): 
            if with_parameter: 
                parameter_stmt = f"and e.parameter = '{parameter}'"
            else: 
                parameter_stmt = ''

            q = f'''
                with models as (
                    select 
                        distinct model_id, 
                        train_end_time, 
                        model_group_id, 
                        model_type, 
                        hyperparameters
                    from triage_metadata.experiment_models join triage_metadata.models using(model_hash)
                    where experiment_hash in ('{"','".join(self.experiment_hashes)}')   
                )
                select 
                    m.model_id, 
                    train_end_time::date as train_end_time_dt,
                    to_char(train_end_time, 'YYYY-MM-DD') as train_end_time,
                    model_type, 
                    model_group_id,
                    stochastic_value as metric_value,
                    parameter
                from models m left join test_results.evaluations e 
                on m.model_id = e.model_id
                and e.metric = '{metric}'
                {parameter_stmt}
                and e.subset_hash = ''
            '''
            return q 
        
        # 1. fetch evaluation values to check if we have 
        q = fetch_evaluation_values()
        df = pd.read_sql(q, self.engine)
        # Validate that we have value for the DEFAULT metric and parameter 
        if df.metric_value.isna().unique(): 
            q = fetch_evaluation_values(with_parameter=False)
            df = pd.read_sql(q, self.engine)
            # fetch the first available value 
            parameter_ = df.loc[0, 'parameter']
            self.threshold = parameter_
            df = df[df.parameter == parameter_]

        df['train_end_time'] = pd.to_datetime(df.train_end_time, format='%Y-%m-%d')
        
        models_per_train_end_time = df.groupby(['model_group_id', 'train_end_time']).count()['model_id']
        
        if models_per_train_end_time[models_per_train_end_time > 1].empty:
            pass 
        else: 
            print(f'model groups with morel than one model id per train end time: \n{models_per_train_end_time[models_per_train_end_time > 1]}' )
        
        if generate_plot:
            # using the audition plotting base here
            plot_cats(
                frame=df,
                x_col='train_end_time',
                y_col='metric_value',
                cat_col='model_type',
                grp_col='model_group_id',
                highlight_grp=None,
                title=f'Model Performance Over Time - {metric}{parameter}',
                x_label='Time',
                y_label=f'Value - {metric}{parameter}',
                cmap_name="tab10",
                figsize=[12, 4],
                dpi=150,
                x_ticks=list(df.train_end_time.unique()),
                y_ticks=None,
                x_lim=None,
                y_lim=(0, 1.1),
                legend_loc=None,
                legend_fontsize=12,
                label_fontsize=12,
                title_fontsize=12,
                label_fcn=None,
                path_to_save=None,
                alpha=0.4,
                colordict=None,
                styledict=None
            )
            sns.despine()     
        
        # sort df by train_end_time
        return df.sort_values(by="train_end_time")
    
    
    def model_performance_subsets(self, metric=None, parameter=None, generate_plot=True):
        
        if metric is None:
            metric = self.performance_metric
        
        if parameter is None: 
            parameter = self.threshold
        
        def fetch_evaluation_values_subsets(with_parameter=True):
            if with_parameter: 
                parameter_stmt = f"and e.parameter = '{parameter}'"
            else: 
                parameter_stmt = ''

            q = f'''
                select 
                    case when e.subset_hash is null then 'full_cohort' 
                    else s.config ->> 'name' 
                    end as "subset",
                    e.subset_hash,
                    m.model_id,
                    m.model_group_id,
                    m.model_type,
                    m.train_end_time::date,
                    e.stochastic_value as metric_value,
                    parameter
                from triage_metadata.experiment_models join triage_metadata.models m using(model_hash)
                    left join test_results.evaluations e
                    on m.model_id = e.model_id
                    and e.parameter = '{parameter}'
                    {parameter_stmt}
                        left join triage_metadata.subsets s on e.subset_hash = s.subset_hash 
                where experiment_hash in ('{"','".join(self.experiment_hashes)}')
            '''

            return q
        
        q = fetch_evaluation_values_subsets() 
        df = pd.read_sql(q, self.engine)
        # Validate that we have value for the DEFAULT metric and parameter 
        if df.metric_value.isna().unique(): 
            q = fetch_evaluation_values_subsets(with_parameter=False)
            df = pd.read_sql(q, self.engine)
            # fetch the first available value 
            parameter_ = df.loc[0, 'parameter']
            self.threshold = parameter_
            df = df[df.parameter == parameter_]
        
        if (df.empty) or (None in df.subset.unique()):
            return None
        
        df['model_type_child'] = df.apply(lambda x: _format_model_name(x['model_type'], x['model_group_id']), axis=1)
        df['model_type_short'] = df.apply(lambda x: x['model_type'].split('.')[-1], axis=1)
        # df.apply(lambda x: _format_model_name(x['model_type']) + ': ' + str(x['model_group_id']), axis=1) 
        
        
        get_labels_table_query = f"""
        select distinct labels_table_name from triage_metadata.triage_runs 
        where run_hash = '{self.experiment_hashes[0]}'
        """

        labels_table = pd.read_sql(get_labels_table_query, self.engine).labels_table_name.iloc[0]

        if generate_plot:        
            grpobj = df.groupby('subset')

            for grp, gdf in grpobj:
                fig, axes = plt.subplots(1, 2, figsize=(10,4), dpi=100)
                fig.suptitle(f'Subset Name: {grp}')
                
                # Plotting the subsetsize and the baserate
                subset_table_name = 'subset_' + gdf['subset'].iloc[0] + '_' + gdf['subset_hash'].iloc[0]
                q = f'''
                    select 
                    as_of_date::date, count(entity_id) as subset_size, avg(label)*100 as baserate
                    from {labels_table} inner join {subset_table_name} using(entity_id, as_of_date)
                    group by 1
                    order by 1
                '''
                subset_size = pd.read_sql(q, self.engine)
                
                # Subset sizes over time
                color='tab:blue'
                sns.barplot(
                    data=subset_size,
                    x='as_of_date',
                    y='subset_size',
                    ax=axes[0],
                    alpha=0.5,
                    color=color
                )
                axes[0].tick_params(axis='x', rotation=90)
                axes[0].set_title(f'Subset Size & Baserate')
                axes[0].set_xlabel('Time')
                axes[0].set_ylabel('Subset size', color=color)
                axes[0].tick_params(axis='y', labelcolor=color)
                axes[0].axhline(y=subset_size.subset_size.mean(), color=color, alpha=0.4, linestyle='--')
                
                
                ax2 = axes[0].twinx()
                color='tab:red'
                
                sns.lineplot(
                    data=subset_size,
                    x=axes[0].get_xticks(),
                    y='baserate',
                    ax=ax2,
                    alpha=0.6,
                    marker='o',
                    markersize=5,
                    color=color
                )
                ax2.set_ylabel('Baserate (%)', color=color)
                ax2.tick_params(axis='y', labelcolor=color)
                ax2.axhline(y=subset_size.baserate.mean(), color=color, alpha=0.4, linestyle='--')
                
                colors={x: sns.color_palette().as_hex()[i] for i, x in enumerate(sorted(df.model_type_short.unique()))}
                legend_handles=[ Line2D([0], [0], label=k, color=v) for k, v in colors.items() ]

                model_groups = gdf.groupby('model_group_id')
                
                for mod_type, model_grp_df in model_groups:
                    model_grp_df = model_grp_df.sort_values('train_end_time')

                    # print(model_grp_df.head())
                    model_type = model_grp_df.model_type_short.iloc[0]
                    
                    model_grp_df.plot(x='train_end_time', y='metric_value', color=colors[model_type], ax=axes[1], alpha=0.4)
                    
                    # ticklabels = list(gdf.train_end_time.unique())
                    # axes[1].set_xticklabels(ticklabels, rotation=45)
                    axes[1].legend(handles=legend_handles, loc='upper left', frameon=False, ncol=1, fontsize='small', bbox_to_anchor=[1, 1])
                    axes[1].tick_params(axis='x', rotation=90)
                    axes[1].set_ylabel(f'{metric}{parameter}')
                    axes[1].set_xlabel('Time')
                    # axes[1].set_ylim(0, 0.3)
                    axes[1].set_title(f'Model Performance')    
                
                plt.tight_layout()

        # sort df by train_end_time     
        return df.sort_values(by="train_end_time")


    def efficiency_and_equity(self, efficiency_metric=None, equity_metric=None, parameter=None, groups=None, model_group_ids=None, bias_metric_tolerance=0.2, generate_plot=True):
        ''' Plot the performanc metric against the bias metric for all or selected models.
            Args:
                
        '''
        
        if efficiency_metric is None:
            efficiency_metric = self.performance_metric
        
        if equity_metric is None:
            equity_metric = self.bias_metric
            
        if parameter is None:
            parameter = self.threshold
            
        if groups is None:
            groups = self.bias_groups

            # check if there are bias audit results in the DB (directly from aequitas table)
            q = f'''
                with attr_values as (
                    select distinct 
                        attribute_name,
                        attribute_value
                    from triage_metadata.experiment_models a 
                    join triage_metadata.models b
                    using(model_hash)
                    join test_results.aequitas c 
                    using (model_id)
                    where experiment_hash = '{self.experiment_hashes[0]}'
                )

                select distinct 
                    attribute_name,
                    attribute_value
                from attr_values
                order by 1,2
            '''
            
            rg = pd.read_sql(q, self.engine)
            
            # if there are no bias audit results in the DB
            if rg.empty:
                logging.warning('No bias audit config or aequitas calculation was not completed! check the test_results.aequitas table. No plots generated')
                return
            else: 
                groups = dict()
                for attr, gdf in rg.groupby('attribute_name'):
                    groups[attr] = list(gdf['attribute_value'].unique())
            
        if model_group_ids is None: 
            # logging.warning('No model groups specified. Usign all model group ids')
            model_group_ids = self.model_groups().model_group_id.tolist()
            
            if not model_group_ids:
                logging.warning('No model groups belong to the experiment! Returning None')
                return None
            
        # If no groups are specified, we show results for all groups    
        if groups is None:
            
            # logging.info('No groups are specified. Showing results for all attributes and their values')
            groups = dict()
            # we can't trust on what is defined in the config, we take the values directly from the aequitas table
            q = f'''
                with attr_values as (
                        select distinct 
                            attribute_name,
                            attribute_value
                        from triage_metadata.experiment_models a 
                        join triage_metadata.models b
                        using(model_hash)
                        join test_results.aequitas c 
                        using (model_id)
                        where experiment_hash = '{self.experiment_hashes[0]}'
                    )

                select distinct 
                    attribute_name,
                    attribute_value
                from attr_values
                order by 1,2
            '''
            
            rg = pd.read_sql(q, self.engine)
            
            if rg.empty:
                logging.warning('No bias audit config or aequitas calculation was not completed! check the test_results.aequitas table. No plots generated')
                return
            
            groups = dict()
            for attr, gdf in rg.groupby('attribute_name'):
                groups[attr] = list(gdf['attribute_value'].unique())
                
            logging.debug(f'Plotting bias for following groups and values {rg}')
            
        
        parameter_ae = parameter
        
        if 'pct' in parameter:
            t = round(float(parameter.split('_')[0]) / 100, 2)
            parameter_ae =  f'{t}_pct'
            
        attributes = groups.keys()
        attribute_values = [v for v_list in groups.values() for v in v_list]
            
        q = f'''
                select 
                    m.model_id,
                    m.model_group_id,
                    m.model_type,
                    m.hyperparameters,
                    m.train_end_time::date,
                    e.metric,
                    e."parameter",
                    e.num_labeled_examples,
                    e.num_positive_labels,
                    e.stochastic_value as "{efficiency_metric}{parameter}",
                    a.tpr,
                    a.{equity_metric},
                    a.attribute_name,
                    a.attribute_value,
                    a.group_size,
                    a.prev as baserate,
                    total_entities
                from triage_metadata.experiment_models em join triage_metadata.models m 
                    on em.model_hash = m.model_hash
                    and m.model_group_id in ({", ".join([str(x) for x in model_group_ids])})
                    left join test_results.evaluations e 
                        on m.model_id = e.model_id 
                        and e.metric = '{efficiency_metric}'
                        and e."parameter" = '{parameter}'
                        and e.subset_hash = ''
                            left join test_results.aequitas a 
                                on m.model_id = a.model_id 
                                and a."parameter" = '{parameter_ae}'
                                and a.attribute_name in ('{"','".join(attributes)}')
                                and a.attribute_value in ('{"','".join(attribute_values)}')
                                and a.tie_breaker= 'worst' 
                where experiment_hash in  ('{"','".join(self.experiment_hashes)}')
                
        '''
        
        #and a.attribute_value in ('{', '.join(attribute_values)}')
        metrics = pd.read_sql(q, self.engine)

        # metrics['Model Class'] = metrics['model_type'].apply(lambda x: x.split('.')[-1])
        # metrics['model_label'] = metrics.apply(lambda x: f"{x['model_group_id']}: {x['model_type'].split('.')[-1]}", axis=1)
        metrics['model_label'] = metrics.apply(lambda x: _format_model_name(x['model_type'], x['model_group_id']), axis=1)
        metrics['model_type_short'] = metrics.apply(lambda x: x['model_type'].split('.')[-1], axis=1)
        
        if generate_plot: 
            # Metric means
            mean = metrics.groupby(['model_label', 'attribute_value', 'model_type_short'])[[f'{efficiency_metric}{parameter}', f'{equity_metric}']].mean().reset_index().sort_values('model_label')
            
            # Metric standard errors
            sem = metrics.groupby(['model_label', 'attribute_value', 'model_type_short'])[[f'{efficiency_metric}{parameter}', f'{equity_metric}']].sem().reset_index().sort_values('model_label')
            labels = sorted(mean.model_label.unique())
            
            # n_attrs = sum([len(x) for x in groups.values()])
            n_attrs = len(attribute_values)
            ax_cntr = 0
            # bias_tolerance = 0.2
            fig, axes = plt.subplots(1, n_attrs, figsize=(4*n_attrs + 1, 4), sharey=True, sharex=True, dpi=100)
            # colors=sns.color_palette().as_hex()[:len(mean.model_label.unique())]
            colors={x: sns.color_palette().as_hex()[i] for i, x in enumerate(sorted(metrics.model_type_short.unique()))}
            legend_handles=[ Line2D([0], [0], label=k, marker='o', color=v) for k, v in colors.items() ]

            
            for group, attrs in groups.items():
                for attr in attrs:
                    msk = mean['attribute_value'] == attr
                    x = mean[msk][f'{efficiency_metric}{parameter}'].tolist()
                    y = mean[msk][f'{equity_metric}'].tolist()

                    # mean[msk]['model_type'].tolist()

                    msk = sem['attribute_value'] == attr
                    yerr = sem[msk][f'{equity_metric}'].tolist()
                    xerr = sem[msk][f'{efficiency_metric}{parameter}'].tolist()

                    # print(x)
                    
                    for i in range(len(x)):
                        axes[ax_cntr].errorbar(x[i], y[i], yerr[i], xerr[i], fmt=' ', linewidth=1, capsize=2, color=colors[mean[msk]['model_type_short'].iloc[i]], alpha=0.3)
                        axes[ax_cntr].scatter(x[i], y[i], color=colors[mean[msk]['model_type_short'].iloc[i]], label=labels[i], alpha=0.5)
                        # axes[ax_cntr].set(title=f'{group} | {attr}', xlabel='Performance Metric', ylabel='Bias Metric', ylim=[0, 3])
                        axes[ax_cntr].set(title=f'{group} | {attr}', xlabel='Efficiency Metric', ylabel='Equity Metric')
                        axes[ax_cntr].axhline(y=1, color='dimgray', linestyle='--', alpha=0.1)
                        axes[ax_cntr].axhline(y=1+bias_metric_tolerance, color='dimgray', linestyle=':', alpha=0.09)
                        axes[ax_cntr].axhline(y=1-bias_metric_tolerance, color='dimgray', linestyle=':', alpha=0.09)
                    ax_cntr += 1
                axes[-1].legend(handles=legend_handles, bbox_to_anchor=(1,1), loc='upper left', frameon=False)
                sns.despine()
            
            # Plotting the Group sizes and baserates
            fig, axes = plt.subplots(1, n_attrs, figsize=(4*n_attrs + 1, 4), sharex=True, sharey=True, dpi=100)
            ax_cntr=0
            for group, attrs in groups.items():
                for attr in attrs:
                    msk = metrics['attribute_value'] == attr

                    grouped = metrics[msk].groupby('train_end_time')[['baserate', 'group_size', 'total_entities']].mean().reset_index()
                    color='tab:blue'
                    sns.barplot(
                        data=grouped,
                        x='train_end_time',
                        y='group_size',
                        ax=axes[ax_cntr],
                        alpha=0.8,
                        color=color,
                        label='Group Size'
                    )

                    sns.barplot(
                        data=grouped,
                        x='train_end_time',
                        y='total_entities',
                        ax=axes[ax_cntr],
                        alpha=0.2,
                        color='gray',
                        label='Cohort Size'
                    )

                    axes[ax_cntr].tick_params(axis='x', rotation=90)
                    axes[ax_cntr].set(xlabel='Time', title=f'{group} | {attr}')
                    axes[ax_cntr].set_ylabel('Entities', color=color)
                    axes[ax_cntr].tick_params(axis='y', labelcolor=color)

                    ax2 = axes[ax_cntr].twinx()
                    color='tab:red'

                    sns.lineplot(
                        data=grouped,
                        x=axes[ax_cntr].get_xticks(),
                        y='baserate',
                        ax=ax2,
                        alpha=0.6,
                        marker='o',
                        markersize=5,
                        color=color,
                        # label='Baserate'
                    )
                    ax2.set_ylabel('Baserate (%)', color=color)
                    ax2.tick_params(axis='y', labelcolor=color)
                    ax_cntr += 1
            axes[-1].legend(bbox_to_anchor=(1.1,1.1), loc='upper left', frameon=False)
            plt.tight_layout()
        
        return metrics
    
      
    def experiment_stats(self):
        q = f'''
            select 
                max(time_splits) as timesplits_from_temporal_config,
                max(matrices_needed/2) as validation_splits,
                max(as_of_times) as as_of_times,
                max(feature_blocks) as feature_groups,
                max(total_features) as features,
                max(matrices_needed) as matrices_needed,
                sum(grid_size) as grid_size,
                sum(models_needed) as models_needed,
                max(case when (matrices_needed/2) < time_splits then 1 else 0 end) as implemented_fewer_splits
            from triage_metadata.experiments
            where experiment_hash in ('{"','".join(self.experiment_hashes)}')
        '''
        
        return pd.read_sql(q, self.engine).to_dict(orient='records')[0]
    
          
    def generate_summary(self, metric=None, parameter=None, equity_metric=None):
        '''
        Summarize:
         - Number of time splits
         - Number of as of dates
         - Average cohort size and baserate
         - Number of features and feature groups
         - Number of model types, model groups, and models built
         - Number of model groups that are missing models 
         - Best model performance and the model group
         - model performance for subsets
         - Best model's bias metric, and whether other models are within the bias metric threshold
        
        '''
        
        if metric is None: 
            metric = self.performance_metric
        
        if parameter is None: 
            parameter = self.threshold
            
        if equity_metric is None:
            equity_metric = self.bias_metric
        
        logger.notice(f"Default performance parameters are set to recall@1_pct and bias metric to tpr_disparity!")
        logger.notice("==> In case your experiment doesn't have those parameters Triage will use one of the available. <==")
        stats = self.experiment_stats()
          
        if stats['implemented_fewer_splits'] == 1:
            logger.notice(f"Temporal config suggests {stats['timesplits_from_temporal_config']} temporal splits, but experiment implemented only {stats['validation_splits']} splits. Was this intentional?")
        else:
            logger.notice(f'Experiment contained {stats["timesplits_from_temporal_config"]} temporal splits')
            
        logger.notice(f"Experiment contained {stats['as_of_times']} distinct as_of_times")
    
        cohorts = self.cohorts(generate_plots=False)
        logger.notice(f'On average, your cohorts contained around {round(cohorts.cohort_size.mean())} entities with a baserate of {round(cohorts.baserate.mean(), 3)}')
        
        logger.notice(f"You built {stats['features']} features organized into {stats['feature_groups']} groups/blocks")
        
        logger.notice(f"Your model grid specification contained {stats['grid_size']} model types with {stats['models_needed']} individual models")
        
        ## Models
        num_models = len(self.models())
        if num_models < stats['models_needed']:
            logger.notice(f"However, the experiment only built {num_models} models. You are missing {stats['models_needed'] - num_models} models")
            
        else:
            logger.notice(f"You successfully built all the {num_models} models")
        
        # Model Performance
        performance = self.model_performance(metric=metric, parameter=parameter, generate_plot=False)
        best_performance = performance.groupby(['model_group_id', 'model_type'])['metric_value'].mean().max()
        best_model_group = performance.groupby(['model_group_id', 'model_type'])['metric_value'].mean().idxmax()[0]
        best_model_type = performance.groupby(['model_group_id', 'model_type'])['metric_value'].mean().idxmax()[1]
        
        # because we could change the value of the default parameter in case it doesn't exist, 
        # it is safer to take it from the object itself.
        logger.notice(f"Your models achieved a best average {self.performance_metric}{self.threshold} of {round(best_performance, 3)} over the {stats['validation_splits']} validation splits, with the Model Group {best_model_group},{best_model_type}. Note that model selection is more nuanced than average predictive performance over time. You could use Audition for model selection.") 
        
        ## Subsets
        subset_performance = self.model_performance_subsets(metric=metric, parameter=parameter, generate_plot=False)
        if subset_performance is not None:
            grpobj = subset_performance.groupby('subset')
            res = []
            for subset, gdf in grpobj:
                d = dict()
                d['subset'] = subset
                d['best_perf'] = round(gdf.groupby(['model_group_id', 'model_type'])['metric_value'].mean().max(),3)
                d['best_mod'] = gdf.groupby(['model_group_id', 'model_type'])['metric_value'].mean().idxmax()

                res.append(d)
            
            if len(res) > 0:
                logger.notice(f"You created {len(res)} subsets of your cohort -- {', '.join([x['subset'] for x in res])}")
                for d in res:
                    logger.notice(f"For subset '{d['subset'] }', Model Group {d['best_mod'][0]}, {d['best_mod'][1]} achieved the best average {metric}{parameter} of {d['best_perf']}")
        else:
            logger.notice("No subsets defined.")

        ## Bias
        equity_metrics = self.efficiency_and_equity(
            efficiency_metric=metric,
            equity_metric=equity_metric,
            parameter=parameter,
            generate_plot=False
        )
        
        if equity_metrics is not None:
            grpobj = equity_metrics[(equity_metrics.baserate > 0) & (equity_metrics.model_group_id == best_model_group)].groupby('attribute_name')
            for attr, gdf in grpobj:
                logger.notice(f"Measuring biases across {attr} groups using {equity_metric} for the best performing model:")
                d = gdf.groupby('attribute_value')[equity_metric].mean()
                logger.notice(", ".join(f"{k}: {round(v, 3)}" for k, v, in d.to_dict().items()))
        else:
            logger.notice(f"No bias audit results were found in the database for the experiment.")
            

    def precision_recall_curves(self, plot_size=(3,3)):
        
        n_splits = self.experiment_stats()['validation_splits']
        n_groups = len(self.model_groups())     
        
        # n_groups x n_splits grid 
        fig, axes = plt.subplots(
            n_groups, 
            n_splits,
            figsize=(n_splits*plot_size[0], n_groups*plot_size[1])
        )
        
        axes = axes.flatten()
        
        grp_obj = self.models().groupby('model_group_id')
        
        ax_idx = 0
        for _, gdf in grp_obj:
            model_ids = gdf.model_id.tolist()
            
            for mod_id in model_ids:
                tmp = SingleModelAnalyzer(engine=self.engine, model_id=mod_id)
                tmp.plot_precision_recall_curve(
                    ax=axes[ax_idx]
                )
                ax_idx += 1
                
            # Making sure that models that aren't built are skipped in the grid
            if ax_idx % n_splits > 0:
                ax_idx += (n_splits - (ax_idx % n_splits)) 

        plt.show()
        plt.tight_layout()


    def get_best_hp_config_for_each_model_type(self):
        
        q = f'''
            with avg_perf as (
                select 
                    model_group_id, model_type, hyperparameters, avg(stochastic_value) as mean_performance
                from triage_metadata.experiment_models join triage_metadata.models m using(model_hash)
                    left join test_results.evaluations e
                    on m.model_id = e.model_id
                    and e.metric = '{self.performance_metric}'
                    and e.parameter = '{self.threshold}'
                    and e.subset_hash = ''
                where experiment_hash in ('{"','".join(self.experiment_hashes)}') 
                and model_type not like '%%Dummy%%'
                group by 1, 2, 3    
            )
            select distinct on(model_type)
            model_group_id, model_type, hyperparameters, mean_performance
            from avg_perf
            order by model_type, mean_performance desc
        '''

        best_models = pd.read_sql(q, self.engine).set_index('model_group_id').sort_values(by='mean_performance', ascending=False)
        best_models['model_type'] = best_models['model_type'].str.split('.').apply(lambda x: x[-1])
            
        return best_models
    
    
    def model_groups_w_best_mean_performance(self, n_model_groups=5):
        """ Return the model groups with the best mean performance """
        
        q = f'''
            with models as (
                select 
                    distinct model_id, train_end_time, model_group_id, model_type, hyperparameters
                from triage_metadata.experiment_models join triage_metadata.models using(model_hash)
                where experiment_hash in ('{"','".join(self.experiment_hashes)}')   
            )
            select 
                m.model_group_id, 
                model_type, 
                hyperparameters,
                avg(stochastic_value) as mean_metric_value
            from models m left join test_results.evaluations e 
                on m.model_id = e.model_id
                and e.metric = '{self.performance_metric}'
                and e.parameter = '{self.threshold}'
                and e.subset_hash = ''
            group by 1, 2, 3
            limit {n_model_groups};
        '''
        
        df = pd.read_sql(q, self.engine)
        
        return df.model_group_id.tolist(), df
    

    def feature_importance(self, plot_size=(2,5), n_features=20):
        n_splits = self.experiment_stats()['validation_splits']
        n_groups = len(self.model_groups())     

        # n_groups x 1 grid 
        fig, axes = plt.subplots(
            n_groups,
            1,
            figsize=(plot_size[0], n_groups*plot_size[1])
        )

        axes = axes.flatten()

        grp_obj = self.models().groupby('model_group_id')

        ax_idx = 0
        for model_group_id, gdf in grp_obj:
            model_ids = gdf.model_id.tolist()

            feature_importances_group = list()
            for mod_id in model_ids:
                tmp = SingleModelAnalyzer(engine=self.engine, model_id=mod_id)

                fi = tmp.get_feature_importances(n_top_features=100)

                feature_importances_group.append(fi)

            feature_importances_group = pd.concat(feature_importances_group)
                        
            if len(feature_importances_group.feature.unique()) > 1: 
                agg_df = feature_importances_group.groupby('feature')['feature_importance'].agg(['mean', 'sem']).nlargest(n_features, 'mean').reset_index()
                
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    sns.barplot(
                        data=agg_df,
                        x='mean',
                        xerr=agg_df['sem'],
                        y='feature',
                        color = 'royalblue',
                        capsize = 0.5,
                        ax = axes[ax_idx]
                    )
                    
                    axes[ax_idx].set_title(f'{_format_model_name(tmp.model_type, model_group_id)}')
                    axes[ax_idx].set_xlabel('')
                    axes[ax_idx].set_ylabel('')
                
            else:
                fig.delaxes(axes[ax_idx])
                logging.warning(f'Not a model with multiple features: {model_group_id} ')
                
            ax_idx += 1

        plt.show()
        plt.tight_layout()
        
    # def feature_group_importance(self):
        


"""empty message

Revision ID: 1b990cbc04e4
Revises: 0bca1ba9706e
Create Date: 2019-02-20 16:41:22.810452

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1b990cbc04e4'
down_revision = '45219f25072b'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("CREATE SCHEMA IF NOT EXISTS production")
    op.execute("ALTER TABLE triage_metadata.list_predictions SET SCHEMA production;")


def downgrade():
    op.execute("ALTER TABLE production.list_predictions SET SCHEMA triage_metadata;")
    op.execute("DROP SCHEMA IF EXISTS production")
